(*
   Title  : FiniteSet_Spec.thy
   Author : M. Bortin
*)

header {* Specification and properties of finite sets *}


theory FiniteSet_Spec
imports Nat
begin

typedecl 'a finite_set


section {* Basic operations on @{typ "'a finite_set"} *}

consts
 Empty     :: "'a finite_set"                         ("{#}")
 Mem       :: "'a \<Rightarrow> 'a finite_set \<Rightarrow> bool"           ("_ <: _" [50, 51] 50) 
 Insert    :: "'a \<Rightarrow> 'a finite_set \<Rightarrow> 'a finite_set"  ("_ ## _" [51, 52] 51) 
 Remove    :: "'a finite_set \<Rightarrow> 'a \<Rightarrow> 'a finite_set"  ("_ -- _" [53, 52] 52)


axioms
 not_Mem_Empty    :   "\<not> (a <: {#})"
 Mem_Insert_eq    :   "a <: b ## S = (a = b \<or> a <: S)"
 Mem_Remove_eq    :   "a <: S -- b = (a \<noteq> b \<and> a <: S)"
 Ext              :   "(\<And> a. (a <: S) = (a <: T)) \<Longrightarrow> S = T"


subsection {* Fold for finite sets *}


consts
 foldSet :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> 'a finite_set \<Rightarrow> 'b"


constdefs
 leftCommuting :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) set"
"leftCommuting \<equiv> {f. \<forall> a b c. f a (f b c) = f b (f a c) }" 

axioms
 foldSet1 : "foldSet f e {#} = e"
 foldSet2 : "f \<in> leftCommuting \<Longrightarrow> \<not> x <: S \<Longrightarrow>
             foldSet f e (x ## S) = f x (foldSet f e S)"




section {* Basic deductions regarding basic operations on finite sets *}


subsection {* Empty set *}

lemma not_Empty_Ex : 
"S \<noteq> {#} \<Longrightarrow> \<exists>a. a <: S"
apply(rule ccontr, erule notE)
apply(rule Ext)
apply(simp add: not_Mem_Empty)
done

lemma not_Empty_Insert :
"a ## S \<noteq> {#}"
apply(rule notI)
apply(drule_tac f="\<lambda>X. a <: X" in arg_cong)
apply(subst (asm) Mem_Insert_eq)
apply(simp add: not_Mem_Empty)
done



subsection {* Insert *}

lemma Insert_fix :
"x <: S \<Longrightarrow> x ## S = S"
apply(rule Ext)
apply(case_tac "a = x", clarify)
apply(subst Mem_Insert_eq, simp)
apply(subst Mem_Insert_eq, simp)
done


corollary Insert_idem :
"a ## (a ## S) = a ## S"
apply(rule Insert_fix)
by (subst Mem_Insert_eq, simp)


lemma Insert_commute :
"a ## (b ## S) = b ## (a ## S)"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(subst Mem_Insert_eq)+
apply simp
apply(case_tac "aa = b", clarify)
apply(subst Mem_Insert_eq)+
apply simp
apply(subst Mem_Insert_eq)+
apply simp
done


subsection {* Remove *}

lemma not_Mem_Remove  :   
"\<not> a <: S -- a"
by(subst Mem_Remove_eq, simp)


lemma Mem_Remove :   
"a \<noteq> b \<Longrightarrow> (a <: S -- b) = (a <: S)"
by(subst Mem_Remove_eq, simp)


lemma Remove_Empty : 
"{#} -- a = {#}"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(simp add:  not_Mem_Remove)
apply(rule not_Mem_Empty)
apply(subst Mem_Remove, assumption)
by(rule refl)


lemma Remove_Insert1 :
"(a ## S) -- a = S -- a"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(simp add: not_Mem_Remove)
apply(subst Mem_Remove, assumption)+
apply(subst Mem_Insert_eq)
by simp

lemma Remove_Insert2 :
"a \<noteq> b \<Longrightarrow> (b ## S) -- a = b ## S -- a"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(simp add: not_Mem_Remove)
apply(subst Mem_Insert_eq)
apply simp
apply(rule not_Mem_Remove)
apply(case_tac "aa = b", clarify)
apply(subst Mem_Remove, assumption)
apply(subst Mem_Insert_eq)+
apply simp
apply(subst Mem_Remove, assumption)
apply(subst Mem_Insert_eq)+
apply simp
apply(subst Mem_Remove, assumption)
by(rule refl)

lemma Remove_commute :
"(S -- a) -- b = (S -- b) -- a"
apply(rule Ext)
apply(case_tac "aa = b", clarify)
apply(case_tac "b = a", clarify)
apply(simp add: not_Mem_Remove)
apply(simp add: Mem_Remove)
apply(rule not_Mem_Remove)
apply(simp add: Mem_Remove)
apply(case_tac "a = aa", clarify)
apply(simp add: not_Mem_Remove)
apply(simp add: Mem_Remove)
done

lemma Insert_Remove :
"a ## S -- a = a ## S"
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(subst Mem_Insert_eq)+
apply simp
apply(subst Mem_Insert_eq)+
apply simp
apply(simp add: Mem_Remove)
done


corollary Insert_Remove_decomp :
"a <: S \<Longrightarrow> S = a ## S -- a"
apply(subst Insert_Remove)
apply(subst Insert_fix, assumption)
by(rule refl)


lemma Remove_fix :
"\<not> a <: S \<Longrightarrow> S -- a = S"
apply(rule Ext)
apply(subst Mem_Remove_eq)
apply force
done


section {* Cardinality *} 

constdefs
 cardinality :: "'a finite_set \<Rightarrow> nat"
"cardinality \<equiv> foldSet (\<lambda>x N. 1 + N) 0"

lemma cardinality0 : "cardinality {#} = 0"
apply(simp add: cardinality_def)
apply(subst foldSet1)
by(rule refl)

lemma cardinality1 : "\<not> a <: S \<Longrightarrow> cardinality (a ## S) = cardinality S + 1"
apply(simp add: cardinality_def)
apply(subst foldSet2)
apply(simp add: leftCommuting_def, assumption)
apply simp
done

lemma cardinality_not_Empty :
"a <: S \<Longrightarrow> 0 < cardinality S"
apply(subst Insert_Remove_decomp, assumption)
apply(subst cardinality1)
apply(rule not_Mem_Remove)
by simp




subsection {* Fold and Remove *}

lemma foldSet3 :
"a <: S \<Longrightarrow> f \<in> leftCommuting \<Longrightarrow> 
foldSet f e S = f a (foldSet f e (S -- a))"
apply(subst Insert_Remove_decomp)
apply(assumption)
apply(subst foldSet2, assumption)
apply(rule not_Mem_Remove)
by(rule refl)

lemma cardinality2 : "a <: S \<Longrightarrow> cardinality (S -- a) = cardinality S - 1"
apply(rule Suc_inject)
apply(frule cardinality_not_Empty)
apply simp
apply(rule sym)
apply(subst cardinality_def)
apply(subst foldSet3)
apply(assumption)
apply(simp add: leftCommuting_def)
apply(fold cardinality_def)
by simp


lemma cardinality_le :
"cardinality (S -- a) \<le> cardinality S" 
apply(case_tac "a <: S")
apply(subst cardinality2, assumption)
apply(simp)
apply(subst Remove_fix, assumption)
by simp


subsection {* Induction *}

lemma Remove_induct[rule_format] :
"P {#} \<Longrightarrow> \<forall>a S. a <: S \<longrightarrow> P (S -- a) \<longrightarrow> P S \<Longrightarrow> P X"
apply(rule_tac f=cardinality in measure_induct)
apply(case_tac "x = {#}", clarify)
apply(drule not_Empty_Ex)
apply clarify
apply(drule spec, drule spec, drule mp, assumption)
apply(erule mp)
apply(drule spec, erule mp)
apply(subst cardinality2, assumption)
apply(drule cardinality_not_Empty)
by simp


lemma Insert_induct[rule_format] :
"P {#} \<Longrightarrow> \<forall>a S. \<not> a <: S \<longrightarrow> P S \<longrightarrow> P (a ## S) \<Longrightarrow> P X"
apply(induct_tac X rule: Remove_induct)
apply assumption
apply(drule_tac x=a in spec)
apply(drule_tac x="S -- a" in spec, drule mp, rule not_Mem_Remove)
apply(drule mp, assumption)
apply(subst (asm) Insert_Remove_decomp[THEN sym], assumption)
by assumption



section {* Finite predicates *}

constdefs
 Repr_FSet :: "'a finite_set \<Rightarrow> 'a set"
"Repr_FSet \<equiv> foldSet insert {}"

 Abst_FSet :: "'a set \<Rightarrow> 'a finite_set"
"Abst_FSet A \<equiv> (THE S. Repr_FSet S = A)"

 FinitePred :: "'a set \<Rightarrow> bool"
"FinitePred A \<equiv> (A \<in> range Repr_FSet)"


lemma insert_leftCommuting[simp] :
"insert \<in> leftCommuting"
by(simp add: leftCommuting_def, force)

lemma Mem_Repr_eq :
"(x \<in> Repr_FSet S) = (x <: S)"  
apply(unfold Repr_FSet_def)
apply(induct_tac S rule: Remove_induct)
apply(subst foldSet1)
apply(simp add: not_Mem_Empty)
apply(subst foldSet3, assumption, simp)
apply simp
apply(case_tac "x = a", simp)
apply simp
apply(subst Mem_Remove, assumption)
by(rule refl)


lemma Repr_FSet_inj :
"inj Repr_FSet"
apply(simp add: inj_on_def)
apply(rule allI)
apply(induct_tac x rule: Remove_induct)
apply clarify
apply(rule Ext)
apply(simp add: not_Mem_Empty)
apply(rule notI)
apply(simp add: Repr_FSet_def)
apply(subst (asm) foldSet1)
apply(subst (asm) foldSet3, assumption, simp)
apply simp
apply(simp (no_asm) add: Repr_FSet_def)
apply(subst foldSet3, assumption, simp)
apply clarify
apply(rename_tac "T")
apply(subgoal_tac "a \<in> Repr_FSet T")
apply(subst (asm) Mem_Repr_eq)
apply(subst (asm) foldSet3, assumption, simp)
apply(fold Repr_FSet_def)
apply(drule_tac x="T -- a" in spec)
apply(drule mp)
apply(rule set_ext)
apply(drule_tac f="\<lambda>X. x \<in> X" in arg_cong) 
apply simp
apply(case_tac "x = a", clarify)
apply(subst Mem_Repr_eq)+
apply(simp (no_asm) add: not_Mem_Remove)
apply simp
apply(rule Ext)
apply(case_tac "aa = a", clarify)
apply(drule_tac f="\<lambda>X. aa <: X" in arg_cong)
apply(subst (asm) Mem_Remove, assumption)+
apply(assumption)
apply force
done

lemma Repr_FSet_inverse :
"Abst_FSet (Repr_FSet S) = S"
apply(unfold Abst_FSet_def)
apply(rule the_equality)
apply(rule refl)
apply(insert Repr_FSet_inj)
by(erule injD, assumption)

lemma Abst_FSet_inverse :
"FinitePred A \<Longrightarrow> Repr_FSet (Abst_FSet A) = A"
apply(unfold FinitePred_def Abst_FSet_def)
apply(simp add: image_def)
apply clarify
apply(rule theI)
apply(rule refl)
apply(insert Repr_FSet_inj)
by(erule injD, assumption)




lemma FiniteEmptyPred :
"FinitePred {}"
apply(unfold FinitePred_def)
apply(simp add: image_def)
apply(rule_tac x="{#}" in exI)
apply(rule set_ext)
apply(subst Mem_Repr_eq)
apply(simp add: not_Mem_Empty)
done


lemma FinitePred_insert :
"FinitePred A \<Longrightarrow> FinitePred (insert a A)"
apply(unfold FinitePred_def)
apply(simp add: image_def)
apply clarify
apply(rule_tac x="a ## x" in exI)
apply(rule set_ext)
apply(subst Mem_Repr_eq)
apply simp
apply(subst Mem_Insert_eq)
apply(subst Mem_Repr_eq)
by(rule refl)

lemma FinitePred_remove :
"FinitePred A \<Longrightarrow> FinitePred (A - {a})"
apply(unfold FinitePred_def)
apply(simp add: image_def)
apply clarify
apply(rule_tac x="x -- a" in exI)
apply(rule set_ext)
apply(subst Mem_Repr_eq)
apply simp
apply(subst Mem_Repr_eq)
apply(subst Mem_Remove_eq)
by force


subsection {* Fold for finite predicates *}

constdefs
 foldPred :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> 'a set \<Rightarrow> 'b"
"foldPred f e \<equiv> foldSet f e \<circ> Abst_FSet"


lemma Abst_FSet_empty :
"Abst_FSet {} = {#}"
apply(rule_tac f=Repr_FSet in injD)
apply(rule Repr_FSet_inj)
apply(subst Abst_FSet_inverse)
apply(rule FiniteEmptyPred)
apply(simp add: Repr_FSet_def)
apply(subst foldSet1)
by(rule refl)


lemma foldPred1 : 
"foldPred f e {} = e"
apply(simp add: foldPred_def)
apply(subst Abst_FSet_empty)
apply(subst foldSet1)
by(rule refl)


lemma Abst_FSet_insert :
"FinitePred A \<Longrightarrow> Abst_FSet (insert x A) = (x ## Abst_FSet A)"
apply(rule_tac f=Repr_FSet in injD)
apply(rule Repr_FSet_inj)
apply(subst Abst_FSet_inverse)
apply(erule FinitePred_insert)
apply(rule set_ext)
apply(subst Mem_Repr_eq)
apply(subst Mem_Insert_eq, simp)
apply(subst Mem_Repr_eq[THEN sym])
apply(subst Abst_FSet_inverse)
apply(assumption)
by(rule refl)


lemma foldPred2 : 
"f \<in> leftCommuting \<Longrightarrow> FinitePred A \<Longrightarrow> \<not> x \<in> A \<Longrightarrow>
 foldPred f e ({x} \<union> A) = f x (foldPred f e A)"
apply(subst foldPred_def)
apply simp
apply(subst Abst_FSet_insert, assumption)
apply(subst foldSet2, assumption)
apply(subst Mem_Repr_eq[THEN sym])
apply(subst Abst_FSet_inverse, assumption)
apply assumption
apply(simp add: foldPred_def)
done


subsection {* Induction on finite predicates *}

lemma Repr_FSet_diff :
"Repr_FSet (S -- a) = Repr_FSet S - {a}"
apply(rule set_ext)
apply(subst Mem_Repr_eq)
apply(subst Mem_Remove_eq)
apply simp
apply(subst Mem_Repr_eq)
by force


lemma FinitePred_induct[rule_format] :
"FinitePred A \<Longrightarrow> P {} \<Longrightarrow> (\<forall>x A. FinitePred A \<longrightarrow> x \<in> A \<longrightarrow> P (A - {x}) \<longrightarrow> P A) \<Longrightarrow> P A"
apply(subst (asm) FinitePred_def)
apply(simp add: image_def)
apply clarify
apply(induct_tac x rule: Remove_induct)
apply(subst Repr_FSet_def)
apply(subst foldSet1)
apply assumption
apply(drule_tac x=a in spec)
apply(drule_tac x="Repr_FSet S" in spec)
apply(drule mp)
apply(simp add: FinitePred_def)
apply(drule mp)
apply(subst Mem_Repr_eq)
apply assumption
apply(erule mp)
apply(subst (asm) Repr_FSet_diff)
by assumption



section {* Union *}


constdefs
 FUnion :: "'a finite_set \<Rightarrow> 'a finite_set \<Rightarrow> 'a finite_set"   ("_ \<squnion> _" [51, 52] 51)
"S \<squnion> T \<equiv> foldSet (\<lambda>s Y. s ## Y) T S" 


lemma FUnion1 :
"a <: S \<Longrightarrow> a <: S \<squnion> T"
apply(drule Insert_Remove_decomp)
apply(erule_tac t=S in ssubst)
apply(subst FUnion_def)
apply(subst foldSet2)
apply(simp add: leftCommuting_def)
apply(subst Insert_commute)
apply(simp)
apply(rule not_Mem_Remove)
apply(simp add: Mem_Insert_eq)
done

lemma FUnion2 :
"a <: T \<Longrightarrow> a <: S \<squnion> T "
apply(induct_tac S rule: Remove_induct)
apply(subst FUnion_def)
apply(subst foldSet1)
apply assumption
apply(subst FUnion_def)
apply(subst foldSet3, assumption)
apply(simp add: leftCommuting_def)
apply(subst Insert_commute)
apply(simp)
apply(fold FUnion_def)
apply(subst Mem_Insert_eq)
by simp



lemma FUnion3[rule_format] :
"a <: S \<squnion> T \<longrightarrow> a <: S \<or> a <: T"
apply(induct_tac S rule: Remove_induct)
apply clarify
apply(subst (asm) FUnion_def)
apply(subst (asm) foldSet1)
apply clarify
apply(subst FUnion_def)
apply(subst foldSet3)
apply assumption
apply(simp add: leftCommuting_def)
apply(subst Insert_commute)
apply(simp (no_asm))
apply(fold FUnion_def)
apply(case_tac "a = aa", clarify)
apply(subst Mem_Insert_eq, simp)
apply(rule impI)
apply(drule mp, assumption)
apply(erule disjE)
apply(subst (asm) Mem_Remove, assumption)
by simp_all


corollary FUnion_eq :
"(a <: S \<squnion> T) = (a <: S \<or> a <: T)"
apply(rule iffI, erule FUnion3)
apply(erule disjE, erule FUnion1, erule FUnion2)
done


lemma FinitePred_Un :
"FinitePred A \<Longrightarrow> FinitePred B \<Longrightarrow> FinitePred (A \<union> B)"
apply(unfold FinitePred_def)
apply(simp add: image_def)
apply clarify
apply(rule_tac x="x \<squnion> xa" in exI)
apply(rule set_ext)
apply(subst Mem_Repr_eq)
apply simp
apply(subst FUnion_eq)
apply(subst Mem_Repr_eq)+
by(rule refl)


lemma Un_Empty : 
"{#} \<squnion> T = T"
apply(rule Ext)
apply(subst FUnion_eq)
apply(simp add: not_Mem_Empty)
done

lemma Un_Insert :
"(x ## S) \<squnion> T = x ## (S \<squnion> T)" 
apply(rule Ext)
apply(subst FUnion_eq)
apply(subst Mem_Insert_eq)+
apply(subst FUnion_eq)
by simp


lemma Un_Remove :
"(S \<squnion> T) -- a = (S -- a) \<squnion> (T -- a)"
apply(rule Ext)
apply(subst FUnion_eq)
apply(subst Mem_Remove_eq)+
apply(subst FUnion_eq)
by force

lemma Un_card_le[rule_format, OF refl] :
" \<forall>S T. S \<squnion> T = X \<longrightarrow> cardinality X \<le> cardinality S + cardinality T"
apply(induct_tac X rule: Remove_induct)
apply clarify
apply(subst cardinality0, simp)
apply clarify
apply(frule Insert_Remove_decomp)
apply(erule ssubst)
apply(subst cardinality1)
apply(rule not_Mem_Remove)
apply(subst (asm) Un_Remove)
apply((drule spec)+, drule mp, rule refl)
apply(rule le_trans)
apply(rule add_le_mono, assumption, rule le_refl)
apply(subst (asm) FUnion_eq)
apply(erule disjE)
apply(subst cardinality2, assumption)
apply(frule cardinality_not_Empty)
apply simp
apply(rule cardinality_le)
apply(subst cardinality2, assumption)
apply(frule cardinality_not_Empty)
apply simp
apply(rule cardinality_le)
done






subsection {* Union of a finite family of finite sets and finite sets as monads *}

constdefs
 FUNION :: "'a finite_set \<Rightarrow> ('a \<Rightarrow> 'b finite_set) \<Rightarrow> 'b finite_set" (infixl ">>=" 5)
"(S >>= i) \<equiv> foldSet (\<lambda>x Y. i x \<squnion> Y) {#} S"

consts
 \<eta> :: "'a \<Rightarrow> 'a finite_set"

defs
 eta_def : "\<eta> \<equiv> (\<lambda>x. x ## {#})"

lemma eta_eq :
"(a <: \<eta> x) = (a = x)"
apply(simp add: eta_def)
apply(subst Mem_Insert_eq)
apply(simp add: not_Mem_Empty)
done


lemma FUnion_leftCommuting[simp] :
"(\<lambda>x. FUnion (f x)) \<in> leftCommuting"
apply(simp add: leftCommuting_def)
apply clarify
apply(rule Ext)
apply(subst FUnion_eq)+
by force


lemma FUNION_EX_eq :
"a <: (X >>= f) = (\<exists>i. i <: X \<and> a <: f i)"  
apply(induct_tac X rule: Remove_induct)
apply(simp add: FUNION_def)
apply(subst foldSet1)
apply(simp add: not_Mem_Empty)
apply(subst FUNION_def)
apply(subst foldSet3)
apply assumption
apply simp
apply(subst FUnion_eq)
apply(subst FUNION_def[symmetric])
apply(rule iffI)
apply(erule disjE)
apply force
apply clarsimp
apply(subst (asm) Mem_Remove_eq) 
apply force
apply (erule exE, erule conjE)
apply (case_tac "i = aa", clarify)
apply(subst (asm) Mem_Remove[THEN sym], assumption)
apply force
done

lemma left_unit :
"(\<eta> x >>= f) = f x"
apply(rule Ext)
apply(subst FUNION_EX_eq)
apply(subst eta_eq)
by force


lemma right_unit :
"(S >>= \<eta>) = S"
apply(rule Ext)
apply(subst FUNION_EX_eq)
apply(subst eta_eq)
by force

lemma FUNION_assoc :  
"(S >>= f >>= g) = (S >>= (\<lambda>x. f x >>= g))"
apply(rule Ext)
apply(subst FUNION_EX_eq)+
by force

lemma eta_inj :
"\<eta> i = \<eta> j  \<Longrightarrow> i = j"
apply(drule_tac f="\<lambda>X. i <: X" in arg_cong)
apply(subst (asm) eta_eq)+
by simp


constdefs
 FMap :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a finite_set \<Rightarrow> 'b finite_set"
"FMap f S \<equiv> (S >>= \<eta> \<circ> f)"

 FJoin :: "('a finite_set) finite_set \<Rightarrow> 'a finite_set"
"FJoin X \<equiv> (X >>= id)" 


lemma FMap_eq :
"(b <: FMap f S) = (\<exists>a. a <: S \<and> b = f a)"
apply(subst FMap_def)
apply(subst FUNION_EX_eq)
apply simp
apply(subst eta_eq)
by(rule refl)

lemma FJoin_eq :
"(a <: FJoin FF) = (\<exists>S. S <: FF \<and> a <: S)" 
apply(subst FJoin_def)
apply(subst FUNION_EX_eq)
apply(simp)
done


constdefs
 FDiff :: "'a finite_set \<Rightarrow> 'a finite_set \<Rightarrow> 'a finite_set" ("_ --- _" [51, 52] 51)
"(S --- T) \<equiv> (S >>= (\<lambda>x. if x <: T then {#} else \<eta> x))"


lemma FDiff_eq :
"(x <: S --- T) = (x <: S \<and> \<not> x <: T)"
apply(unfold FDiff_def)
apply(subst FUNION_EX_eq)
apply(rule iffI)
apply clarify
apply(split split_if_asm)
apply(simp add: not_Mem_Empty)
apply(subst (asm) eta_eq, clarsimp)
apply(rule_tac x=x in exI, simp)
apply(subst eta_eq)
by(rule refl)

lemma FinitePred_Diff :
"FinitePred A \<Longrightarrow> FinitePred B \<Longrightarrow> FinitePred (A - B)"
apply(unfold FinitePred_def)
apply(simp add: image_def)
apply clarify
apply(rule_tac x="x --- xa" in exI)
apply(rule set_ext)
apply(subst Mem_Repr_eq)
apply simp
apply(subst FDiff_eq)
apply(subst Mem_Repr_eq)+
by(rule refl)


constdefs
 FInter :: "'a finite_set \<Rightarrow> 'a finite_set \<Rightarrow> 'a finite_set" ("_ \<sqinter> _" [51, 52] 51)
"(S \<sqinter> T) \<equiv> S --- (S --- T)" 


lemma FInter_eq :
"(x <: S \<sqinter> T) = (x <: S \<and> x <: T)"
apply(unfold FInter_def)
apply(subst FDiff_eq)+
apply force
done


lemma FinitePred_Inter :
"FinitePred A \<Longrightarrow> FinitePred B \<Longrightarrow> FinitePred (A \<inter> B)"
apply(unfold FinitePred_def)
apply(simp add: image_def)
apply clarify
apply(rule_tac x="x \<sqinter> xa" in exI)
apply(rule set_ext)
apply(subst Mem_Repr_eq)
apply simp
apply(subst FInter_eq)
apply(subst Mem_Repr_eq)+
by(rule refl)


end