(* **************************************************************************
 *
 * Various auxiliary functions.
 *
 * $Id: auxiliary.sml 1190 2009-12-07 13:51:33Z maxim $
 *)

structure Aux =

struct

(* ************************************************************* *)
(*                    Auxiliaries                                *)
(* ************************************************************* *)


(* ------------------------------------------------------------------------- *)
(* common functions for argument parsing/scanning                            *)
(* ------------------------------------------------------------------------- *)

(* mapping *)
val scan_mapsto = ((OuterParse.$$$ "|->") || (OuterParse.$$$ "\<mapsto>"))
val scan_assign = (fn ((a, b), t) => (([a], b), t)) o (OuterParse.string -- (scan_mapsto |-- OuterParse.string))
val scan_assign_comp = ( OuterParse.$$$ "(" |-- (OuterParse.list1 OuterParse.string --  
                       (scan_mapsto |-- (OuterParse.string --| OuterParse.$$$ ")"))))
val scan_mapping = (OuterParse.$$$ "[" |-- (OuterParse.list1 (scan_assign || scan_assign_comp)) 
                --| OuterParse.$$$ "]") 

(* renaming *)            
val scan_mix = Scan.option (((OuterParse.$$$ "mixfix") -- (OuterParse.$$$ ":")) |-- 
                            (OuterParse.opt_mixfix || OuterParse.opt_infix))
val scan_ren_assign      = (OuterParse.$$$ "(" |-- (OuterParse.string --  
                                  (scan_mapsto |-- ((OuterParse.name -- scan_mix) --| 
                                  OuterParse.$$$ ")"))))
val scan_renaming = (OuterParse.$$$ "[" |-- (OuterParse.list1 scan_ren_assign) --| OuterParse.$$$ "]");



(* ****************** *)
(* *** Lazy lists *** *)
(* ****************** *)

datatype 'a LList = LNil | LCons of ('a * (unit -> 'a LList))

fun force lx = lx ()
fun lapp LNil ll = force ll
  | lapp (LCons (x, f)) ll = LCons(x, (fn () => (lapp (force f) ll)))
fun lmap f LNil = LNil
  | lmap f (LCons(x, lxs)) = LCons(f x, fn () => lmap f (force lxs))
fun map_to_list f LNil = []
  | map_to_list f (LCons(x, lxs)) = (f x)::(map_to_list f (force lxs))
        

fun lprod f lxs lys =
 let fun f1 x LNil = LNil
       | f1 x (LCons (y, ll)) = LCons((f x, f y), fn () => f1 x (force ll))
     fun f2 LNil ys = LNil
       | f2 (LCons (x, ll)) ys = lapp (f1 x ys) (fn () => f2 (force ll) ys)
 in
  f2 lxs lys
 end

(* ********** *)



(* output *)

fun writeln s = Output.writeln s
fun warning s = Output.warning s
fun error_msg s = Output.error_msg s
fun error s = (error_msg s; raise ERROR s);
fun info s = priority s 


(*association list lookup, optimized version for strings*)
fun assoc_string ([], (key:string)) = NONE
  | assoc_string ((keyi, xi) :: pairs, key) =
      if key = keyi then SOME xi else assoc_string (pairs, key);

fun assoc_string2 ([], (key:string)) = NONE
  | assoc_string2 ((xi, keyi) :: pairs, key) =
      if key = keyi then SOME xi else assoc_string2 (pairs, key);

fun assoc_int ([], (key:int)) = NONE
  | assoc_int ((keyi, xi) :: pairs, key) =
      if key = keyi then SOME xi else assoc_int (pairs, key);

fun foldl f (e, []) = e
  | foldl f (e, x::xs) = foldl f (f (e, x), xs)
fun foldr f ([], e) = e
  | foldr f (x::xs, e) = f(x, foldr f (xs, e))

fun map_foldl g f (e, []) = e
  | map_foldl g f (e, x :: xs) = map_foldl g f (f (e, g x), xs)

fun flat (ls: 'c list list) : 'c list = foldl (op @) ([], ls);

fun overwrite (al, p as (key, _)) =
  let fun over ((q as (keyi, _)) :: pairs) =
            if keyi = key then p :: pairs else q :: (over pairs)
        | over [] = [p]
  in over al end;

fun mapfilter f [] = []
  | mapfilter f (x::xs) = (case (f x) of 
                             NONE => mapfilter f xs
                           | SOME y => y::(mapfilter f xs))
fun mapfilter2 P f [] = []
  | mapfilter2 P f (x::xs) = (if P x then (f x)::(mapfilter2 P f xs)
                              else mapfilter2 P f xs)


fun seq (proc: 'a -> unit) : 'a list -> unit =
  let fun seqf [] = ()
        | seqf (x :: xs) = (proc x; seqf xs)
  in seqf end;


fun extend_symtab (table, args) = fold Symtab.update_new args table;



(* *** Thms and axioms lookups *** *)

val all_thms_of = PureThy.all_thms_of
val get_axiom = Thm.axiom

fun thms_of thy =
 let val pr = Theory.parents_of thy
     val fs = map PureThy.facts_of pr
 in
  Facts.dest_static fs (PureThy.facts_of thy)
 end

fun flat_thms_of thy =
 let fun fl nm i [] = []
       | fl nm i (x::xs) = (nm^"_"^(string_of_int i), x)::(fl nm (i + 1) xs)
     fun flat [] = []
       | flat ((nm, t)::xs) = (case t of
                                [x] => (nm, x)::(flat xs)
                               | _ => (fl nm 1 t)@(flat xs))
 in
  flat (thms_of thy)
 end 


(* ***** Operations and logical types of a signature ***** *)

fun sort_alg thy = (let val tsig = Sign.tsig_of thy
                        val cls = #classes (Type.rep_tsig tsig)
                    in
                      snd cls
                    end)

fun all_classes thy = (let val alg = sort_alg thy
                       in
                         Sorts.all_classes alg
                       end)

fun is_abbreviation_ T c = (let val x = SOME (Consts.the_abbreviation T c)
                                         handle TYPE _ => NONE 
                             in
                              case x of
                               NONE => false
                             | _ => true
                             end)
                              

fun get_ops sg = (let val T = #consts (Sign.rep_sg sg)
                      val c = #constants(Consts.dest T)
                  in
                   filter (not o (is_abbreviation_ T)) (Symtab.keys (snd c)) 
                    (*  (Symtab.keys (snd c)) *)
                  end)

fun get_ltypes sg = Type.logical_types (Sign.tsig_of sg)

fun const_type thy c = (case (Sign.const_type thy c) of
                         SOME ty => SOME ty
                       | NONE => 
                       (let val t = snd (#constants(Consts.dest (#consts (Sign.rep_sg thy))))
                        in
                         case (Symtab.lookup t c) of
                          NONE => NONE
                        | SOME (ty, _) => SOME ty
                        end))

fun type_arity(sg, tycon) = 
 let val alg = sort_alg sg
     val t = #arities (Sorts.rep_algebra alg)
 in
  case (Symtab.lookup t tycon) of
   SOME a => (case a of 
               [] => NONE
             | _ =>  (let val a' = map snd a
                      in
                       SOME (a', length (snd(hd a')))
                      end))
 | NONE => NONE
 end


fun type_rank thy tycon = 
 let val tsig = Sign.tsig_of thy
 in
  Type.arity_number tsig tycon
 end

val varifyT = Term.map_atyps (fn TFree (a, S) => TVar ((a, 0), S) | T => T);
val varify =
  Term.map_aterms (fn Free (x, T) => Var ((x, 0), T) | t => t) #>
  Term.map_types varifyT;


fun unvarifyT (TVar ((v, i), s)) = TFree (v, s)
  | unvarifyT (Type (f, ts)) = Type (f, map unvarifyT ts)
  | unvarifyT t = t 

fun unvarify (Const (a, T)) = Const (a, unvarifyT T)
  | unvarify (Free (a, T)) = Free (a, unvarifyT T)
  | unvarify (Var ((a, x), T)) = Free (a, unvarifyT T)
 (* | unvarify (Var (ixn, T)) = Var (ixn, unvarifyT T) *) (*non-0 index!*)
  | unvarify (t as Bound _) = t
  | unvarify (Abs (a, T, body)) = Abs (a, unvarifyT T, unvarify body)
  | unvarify (f $ t) = unvarify f $ unvarify t;



(* ******************************************************* *)


(* *** Identifiers *** *)

fun is_dot #"." = true
  | is_dot   _  = false

fun has_dot s = (let val l = explode s
                     fun chk [] = false
                       | chk (x::xs) = (x = ".") orelse (chk xs)
                 in
                  chk l
                 end)


val base = Long_Name.base_name
val qualifier = Long_Name.qualifier
val is_qualified = Long_Name.is_qualified
val is_hidden = String.isPrefix "??.";

val pack = space_implode "."
val unpack = space_explode "."
val implode_path = Long_Name.implode
val explode_path = Long_Name.explode

fun qname_split qnm =
 let val sp = unpack qnm 
 in
  case sp of
   [x] => ("", x)
 | q::t => (q, space_implode "_" t)
 | _ => error ("Illegal name " ^ qnm)
 end      

fun qname_split2 qnm =
 let val sp = unpack qnm 
 in
  case sp of
   [x] => ("", x)
 | q::t => (q, space_implode "." t)
 | _ => error ("Illegal name " ^ qnm)
 end    

fun path_base p =
 let val sp = space_explode "/" p 
 in
  hd(unpack(List.last sp))
 end



(* *** Theory and thm names  *** *)

fun name_of_thy thy = Context.theory_name thy

fun name_of_thm t = 
(let val tgs = (#tags o Thm.rep_thm) t
 in
  assoc_string (tgs, "name")
 end)




(* ********* ThyMorph Auxiliaries ************ *) 

fun show_opmap (s, (snm, tnm)) = (s ^ "\nFound: (\"" ^ snm ^ "\" |-> \"" ^ tnm ^ "\")" )
fun show_amb (s, (nm, nmlist)) = (s ^ "\n\"" ^ nm ^ "\" |-> [ " ^ (commas nmlist) ^ " ]")

fun gen_arity 0 = []
  | gen_arity n = (let val c = ("'" ^ (chr (96 + n)))
                   in
                    c :: (gen_arity (n - 1))
                   end)


fun check_sko_ c = (case (try (unprefix "sko_") c) of
                          NONE => false
                        | _ => true)
fun check_random_ c = (case (try (unprefix "random_") c) of
                             NONE => false
                           | _ => true)

fun check_typerep_ c = (case (try (unprefix "typerep_") c) of
                              NONE => false
                            | _ => true)

fun check_term_of_ c = (case (try (unprefix "term_of_") c) of
                              NONE => false
                            | _ => true)

fun check_ c = (check_sko_ c) orelse (check_random_ c) orelse (check_typerep_ c) orelse (check_term_of_ c)

fun internal_id c =
 let val (a, b) = qname_split2 c
 in
  ((check_ (base c)) orelse
  (if (a = "") then false
   else (let val (a', b') = qname_split2 b
         in
          not (a' = "")
         end))) 
 end


(* Getting dependencies *)


fun ins_string (x : string) = insert (op =) x 
fun union_str (a : string list, b) = union (op =) a b
fun union_strs xs = foldl union_str ([], xs)
fun diff_str (a : string list, b) = subtract (op =) b a 
fun inter_str (a : string list, b) = inter (op =) a b


fun add_term_consts (Const (c, _), cs) = insert (op =) c cs
  | add_term_consts (t $ u, cs) = add_term_consts (t, add_term_consts (u, cs))
  | add_term_consts (Abs (_, _, t), cs) = add_term_consts (t, cs)
  | add_term_consts (_, cs) = cs

fun term_consts t = add_term_consts(t,[])

fun all_consts_of_prop thm = term_consts (prop_of thm)

fun all_tycons_of_type (Type (f, ts)) = ins_string f (union_strs (map all_tycons_of_type ts))
  | all_tycons_of_type _ = []

fun all_classes_of_type (Type (f, ts)) = (union_strs (map all_classes_of_type ts))
  | all_classes_of_type (TFree (x, s)) = s
  | all_classes_of_type (TVar (i, s)) = s

fun deps_of_term thy t =
 let val consts = term_consts t
     val types = map (the o (const_type thy)) consts
     val tycons = union_strs (map all_tycons_of_type types)
     val classes = union_strs (map all_classes_of_type types)
 in
  (classes, tycons, consts)
 end

fun all_classes_of_tycon thy tc =
 let val ar = (case type_arity(thy, tc) of
                NONE => [] (* error ("Unknown type constructor " ^ tc ^ " in all_classes_of_tycon") *)
              | SOME x => fst x)
     fun f [] = []
       | f ((c, sl)::xs) = union_str(ins_string c (union_strs sl), f xs)
 in
  f ar
 end 
 
fun get_axclass_info thy x = SOME (AxClass.get_info thy x)
                             handle e => NONE


(* ******************* *)

val rhs = snd o Logic.dest_equals
          handle TERM _ => error "wrong argument for rhs"

fun check_overload alg typ term = 
 let val typ' = Term.fastype_of term
     fun chk_arg (Type (f, ts), Type (g, ts')) = ((f = g) andalso (forall chk_arg (ts ~~ ts')))
       | chk_arg (TFree (x, s), TFree (y, s')) = (Sorts.sort_eq alg (s, s'))
       | chk_arg  _ = false
     fun check_types (Type("fun", [l, r])) (Type("fun", [l', r'])) = 
           (if (chk_arg (l, l')) then check_types r r' else true)
       | check_types ty ty' = not (chk_arg (ty, ty'))
 in
  check_types typ typ'
 end

fun cert_def thy thm = (SOME (Sign.cert_def (ProofContext.init thy) ((unvarify o prop_of) thm))
                        handle e => NONE)


(* For unvarified def. terms *)
fun abs_def_i (Const ("==>", t) $ A $ B) = (Const("==>", t) $ A $ (abs_def_i B))
  | abs_def_i term = Logic.mk_equals (Primitive_Defs.abs_def term)
                     handle TERM _ => term
val abs_def = abs_def_i o unvarify  




infix 7 >>=
fun (opt >>= f) = case opt of
                    NONE => NONE
                  | SOME x => f x

fun get_opname (Const ("==>", t) $ A $ B) = get_opname B
  | get_opname tm = 
    (try Logic.dest_equals) (Logic.strip_imp_concl tm) >>= 
    (fn (lhs, rhs) => SOME (Term.strip_comb lhs)) >>= 
    (fn (head, args) => (try Term.dest_Const) head) >>=
    (fn (c, T) => SOME c) 


fun set_opname1 (Const(nm, ty)) nnm = Const(nnm, ty)
  | set_opname1 (s $ t) nnm = (set_opname1 s nnm) $ t
  | set_opname1 _ nnm = error "Unexpected pattern in set_opname1" 
fun set_opname (Const("==>", ty) $ s $ t) nnm = (Const("==>", ty) $ s $ (set_opname t nnm))
  | set_opname (Const("==", ty) $ s $ t) nnm = (Const("==", ty) $ (set_opname1 s nnm)) $ t
  | set_opname (s $ t) nnm = (set_opname s nnm) $ t
  | set_opname _ nnm = error "Unexpected pattern in set_opname"



fun unq_path from to =
 let val fname = name_of_thy from
     fun path p thy = (if ((name_of_thy thy) = fname) then p
                       else (case (Theory.parents_of thy) of
                              [thy'] => (path (thy::p) thy')  
                            | _ => []))
 in
  path [] to
 end 



val empty_list = Const ("List.list.Nil", Type ("List.list", [Type ("List.char", [])]))

fun norm_STR (Const("String.literal.STR", ty) $ t) = 
             (Const("String.literal.STR", ty) $ empty_list)
  | norm_STR (s $ t) = (norm_STR s) $ (norm_STR t)
  | norm_STR (Abs(x, ty, t)) = (Abs(x, ty, norm_STR t))
  | norm_STR t = t



(* ************************************************************* *)
(*           PROOF RECONSTRUCTION AND THEOREM DERIVATION         *)
(* ************************************************************* *)


fun get_thm thy =
  let val tab = fold_rev Symtab.update (PureThy.all_thms_of thy) Symtab.empty
  in
    (fn s => case Symtab.lookup tab s of
       NONE => error ("Unknown theorem " ^ quote s)
     | SOME thm => thm)
  end


fun get_thm' thy s = 
    (case (assoc_string (PureThy.all_thms_of thy, s)) of 
	 NONE => error ("Unknown theorem " ^ quote s)
       | SOME thm => thm)

(* Freezes free variables in proof *)
fun freeze prf = 
      let val (pt2, _)= Proofterm.freeze_thaw_prf (prf) in pt2 end


(* Strip a PThm followed by applications of %;
 * ie. strip_thm PThm(..., prf, ...) % .. % ..  is prf.
 * Note that Proofterm.strip_thm also strips away outer %%, which seems
 * potentially harmful for our uses.
 *)
fun strip_thm prf = 
   (case Proofterm.strip_combt prf of
       (PThm (_, (_, f)), _) => Proofterm.join_proof f
      | _ => prf)


(* Accessor functions on theorem representations *)
fun get_prf thm  =  Thm.proof_of thm (* Does not work with Isa2009 #2 (#der (rep_thm thm)) *)
(* ... and strip top Prf(..) *)
fun proof_of thm = strip_thm (get_prf thm)

fun proof_of' thm = (Thm.get_name thm, proof_of thm)


(* Checks whether a theorem "name" occurs in a proof *)
local open Proofterm
in

fun occurs_thm name (PThm (_, ((thmnm, _, _), _))) = (name = thmnm)
  | occurs_thm name (Abst(_, _, p)) = occurs_thm name p
  | occurs_thm name (AbsP(_, _, p)) = occurs_thm name p
  | occurs_thm name (p % _)         = occurs_thm name p
  | occurs_thm name (p %% q)        = ((occurs_thm name p) orelse (occurs_thm name q))
  | occurs_thm _    _               = false
 


fun all_occurs_thm_ nms (PThm(_, ((thmnm, _, _), _))) = 
    (if ((thmnm = "") orelse (thmnm mem_string nms)) then nms else (thmnm::nms))
  | all_occurs_thm_ nms (Abst(_, _, p)) = all_occurs_thm_ nms p
  | all_occurs_thm_ nms (AbsP(_, _, p)) = all_occurs_thm_ nms p
  | all_occurs_thm_ nms (p % _)         = all_occurs_thm_ nms p
  | all_occurs_thm_ nms (p %% q)        = (let val nms1 = all_occurs_thm_ nms p 
                                           in
                                            all_occurs_thm_ nms1 q
                                           end)
  | all_occurs_thm_ nms _ = nms

end

val all_occurs_thm = all_occurs_thm_ []




(* **************************************************** *
 * Necessary to expand reconstructed proofterm for      *
 * anonymous theorems (""), this is now done in one go  *
 * by functions "reconstruct_prf" and "fullprf_of".     *
 * **************************************************** *)

(* ISABELLE2003: modified argument to     
 * expand_proof from [""] to [("", NONE)] *)
fun reconstruct_prf sg prop prf= 
   let val prf' = Reconstruct.reconstruct_proof sg prop prf
   in if (occurs_thm "" prf')
      then (Reconstruct.expand_proof sg [("", NONE)] prf') 
      else prf' 
   end



fun reconstruct_prf' sg prop nm prf = 
let val prf' = Reconstruct.reconstruct_proof sg prop prf
    val prf'' = (if (occurs_thm nm prf')
                  then (Reconstruct.expand_proof sg [(nm, NONE)] prf')
                  else prf')
in
  if (occurs_thm "" prf'') then (Reconstruct.expand_proof sg [("", NONE)] prf'')
                           else prf''  
end


fun fullprf_of thm = reconstruct_prf' (Thm.theory_of_thm thm) (prop_of thm) (Thm.get_name thm) (proof_of thm)

fun fullprf_of' thm = 
      (Thm.get_name thm, fullprf_of thm)


(* *********************************** *)
(* ***         derive_thm          *** *)
(* *********************************** *) 

val beta_eta_convert = Conv.fconv_rule Drule.beta_eta_conversion


(* Derive theorem from proofterm in given theory *)
fun derive_thm thy prf = 
     Drule.standard (Thm.varifyT (ProofChecker.thm_of_proof thy (freeze prf))) 
        handle e=> let val _ = writeln ("Exception in derive_thm: ")
                    in raise e end;



(* Verify that the proof of a theorem can be proofchecked *)
fun check_prf_of thy thm = derive_thm thy (fullprf_of thm)



end
