(* 
 * Signature morphisms
 * $Id: sigmorph.sml 1194 2010-03-01 15:04:35Z maxim $
 *)


signature SIGMORPH =
sig
  type sigmorph

  datatype SigMorphData = SigMorphData of 
   { dom : string list,
     cod : string list,
     global : string list,
     class_inter : Term.class list,
     class_basis : Term.class list,
     type_inter : string list,
     op_inter   : string list,
     type_basis : string list,
     op_basis   : (string * Term.typ) list,
     der_ops    : (string * Term.typ) list,
     dom_defnm_rel  : (string list) Symtab.table,  (* maps any name from dom_der_ops to its defs *)
     cod_defnm_rel  : (string list) Symtab.table,
     dom_def_map : ((string * Term.term) * Symtab.key list) Symtab.table,
     dom_axs_map : Term.term Symtab.table,
     cod_def_map : ((string * Term.term) * Symtab.key list) Symtab.table,
     cod_axs_map : Term.term Symtab.table,  
     sorted_dom_defs : (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list,
     sorted_cod_defs : (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list
   }
  val get_sigmorph_data : theory -> theory -> SigMorphData
  val src : sigmorph -> theory
  val trg : sigmorph -> theory
  val classMap : sigmorph -> Term.class Symtab.table
  val typeMap : sigmorph -> TypeS.TS Symtab.table
  val opMap   : sigmorph -> Symtab.key Symtab.table
  val defMap : sigmorph -> Symtab.key Symtab.table
  val data    : sigmorph -> SigMorphData
  val map_sort : sigmorph -> Term.sort -> Term.sort
  val map_type   : sigmorph -> Term.typ -> Term.typ
  val map_term   : sigmorph -> Term.term -> Term.term
  val map_proof   : sigmorph -> Proofterm.proof -> Proofterm.proof
  val maps_class : sigmorph -> Term.class -> Term.class
  val maps_tycon  : sigmorph -> string -> TypeS.TS
  val maps_const  : sigmorph -> string -> string 
  val match_defs : sigmorph -> sigmorph
  val normalize : (string * mixfix option) Symtab.table -> sigmorph -> sigmorph 
  val compose : (string * mixfix option) Symtab.table -> sigmorph -> sigmorph -> sigmorph * sigmorph
  val cod_ext : sigmorph -> theory -> sigmorph 
  val translate_syntax : (string * mixfix option) Symtab.table -> sigmorph -> sigmorph


  datatype Sigmorph_result = Just of sigmorph
                           | AmbiguousConsts of (string * string list) list * (string * string) list * 
                                                (Term.sort -> Term.sort) * (Term.typ -> Term.typ) 

  val sigmorph_cmd : theory -> theory -> (string * string) list -> 
                     (string * string) list -> (string * string) list -> Sigmorph_result   
  val make : theory -> theory -> (string * string) list ->
             (string * TypeS.TS) list -> (string * string) list -> Sigmorph_result   
  val instantiate : bool -> (string * string) list -> (string * string) list -> (string * string) list -> 
                   (string * mixfix option) Symtab.table -> 
                    theory -> sigmorph -> (sigmorph * Derive_Thm.AxmImg Symtab.table)


  val add_sigmorph    : (string * sigmorph) -> unit
  val get_sigmorph    : string -> sigmorph option
  val print_sigmorphs : unit -> string
  val verbose : bool Unsynchronized.ref

  val sigmorph_to_string : sigmorph -> string

  exception SIGMORPH of string

end


structure Sigmorph : SIGMORPH =
struct


val verbose = Unsynchronized.ref false 
fun print_verb s = if !verbose then writeln s else () 

  datatype SigMorphData = SigMorphData of 
   { dom : string list,
     cod : string list,
     global : string list,
     class_inter : Term.class list,
     class_basis : Term.class list,
     type_inter : string list,
     op_inter   : string list,
     type_basis : string list,
     op_basis   : (string * Term.typ) list,
     der_ops    : (string * Term.typ) list,
     dom_defnm_rel  : (string list) Symtab.table,  (* maps any name from dom_der_ops to its defs *)
     cod_defnm_rel  : (string list) Symtab.table,
     dom_def_map : ((string * Term.term) * Symtab.key list) Symtab.table,
     dom_axs_map : Term.term Symtab.table,
     cod_def_map : ((string * Term.term) * Symtab.key list) Symtab.table,
     cod_axs_map : Term.term Symtab.table,  
     sorted_dom_defs : (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list,
     sorted_cod_defs : (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list
   }



   datatype sigmorph = SigMorph of { src : theory,
		         	     trg : theory,
                                     classMap : Term.class Symtab.table,
				     typeMap : TypeS.TS Symtab.table,
				     opMap   : Symtab.key Symtab.table,
                                     defMap : Symtab.key Symtab.table,
                                     data : SigMorphData
				    }

fun src (SigMorph{src, ...}) = src
fun trg (SigMorph{trg, ...}) = trg
fun classMap (SigMorph{classMap=t, ...}) = t
fun typeMap (SigMorph{typeMap=t, ...}) = t
fun opMap (SigMorph{opMap=t, ...}) = t
fun defMap (SigMorph{defMap=t, ...}) = t
fun data (SigMorph{data=d, ...}) = d


exception SIGMORPH of string


(* auxiliaries *)

     fun assoc_defs defnm_rel dep_ops = Aux.union_strs (map (Symtab.lookup_list defnm_rel) dep_ops)
     fun new_name names pre id = (if (id mem_string names) then new_name names pre (pre ^ "_" ^ id)
                                  else id) 

     fun map_option f [] = []
       | map_option f (x::xs) = (case (f x) of
                                  NONE => x::(map_option f xs)
                                | SOME y => y::(map_option f xs))

     fun merge_list rel e_rel =
      (let val t = Symtab.dest_list e_rel
       in
         Aux.foldr (uncurry Symtab.cons_list) (t, rel)
       end)


     fun ext_mapping mp [] = mp
       | ext_mapping mp ((x as (key, v))::xs) =
         let val mp' = Symtab.update_new x mp
                       handle DUP => (let val _ = writeln ("Updating: " ^ key ^ " |-> " ^ v)
                                      in
                                       Symtab.update x mp
                                      end)
         in
          ext_mapping mp' xs
         end


     fun split_axms thy (axs, defnm_rel, defs) [] = (axs, defnm_rel, defs)
       | split_axms thy (axs, defnm_rel, defs) ((nm, thm)::xs) =
         (case (Aux.cert_def thy thm) of
             SOME ((id, ty), rhs) => (let val _ = (if !verbose then Aux.info ("Definition " ^ nm ^ " found")
                                                   else ())
                                      in
                                      split_axms thy (axs, Symtab.cons_list (id, nm) defnm_rel, 
                                                     (nm, id, thm)::defs) xs
                                      end)
           | NONE => (let val _ = (if !verbose then Aux.info ("Proper axiom " ^ nm ^ " found")
                                   else ())
                      in
                       split_axms thy ((nm, thm)::axs, defnm_rel, defs) xs
                      end))

     fun select_der_metaeq thy (defnm_rel, defs) [] = (defnm_rel, defs)
       | select_der_metaeq thy (defnm_rel, defs) ((nm, thm)::xs) = 
         (case (Aux.cert_def thy thm) of
             SOME ((id, ty), rhs) => 
              (case (Symtab.lookup defnm_rel id) of 
                NONE => select_der_metaeq thy (Symtab.cons_list (id, nm) defnm_rel, (nm, id, thm)::defs) xs
              | SOME ds => (if (nm mem_string ds) 
                            then select_der_metaeq thy (defnm_rel, defs) xs
                            else select_der_metaeq thy (Symtab.cons_list (id, nm) defnm_rel, 
                                                        (nm, id, thm)::defs) xs))
           | NONE => select_der_metaeq thy (defnm_rel, defs) xs)


     fun prepare_defs thy defnm_rel t [] = t
       | prepare_defs thy defnm_rel t ((nm, id, thm)::xs) =
         let val dep = Aux.all_consts_of_prop thm  
             val dep' = Aux.diff_str (assoc_defs defnm_rel dep, [nm])
             val norm = Aux.abs_def_i o Aux.unvarify o prop_of o Aux.beta_eta_convert
             val norm_STR = Aux.abs_def_i o Aux.unvarify o Aux.norm_STR o prop_of o Aux.beta_eta_convert 
         in
          if (Aux.internal_id id)
          then prepare_defs thy defnm_rel (Symtab.update (nm, ((id, norm_STR thm), dep')) t) xs
          else prepare_defs thy defnm_rel (Symtab.update (nm, ((id, norm thm), dep')) t) xs
         end 

     fun prepare_axs t [] = t
       | prepare_axs t ((nm, thm)::xs) =
         let val norm = Aux.unvarify o prop_of o Aux.beta_eta_convert
         in
          prepare_axs (Symtab.update (nm, norm thm) t) xs
         end 
   
 
      fun split_ops thy t (log, der) [] = (log, der)
        | split_ops thy t (log, der) (c::cs) =
          (case (Symtab.lookup_list t c) of
            [] => (case (Aux.const_type thy c) of
                     NONE => split_ops thy t ((c, dummyT)::log, der) cs
                   | SOME ty => split_ops thy t ((c, Aux.unvarifyT ty)::log, der) cs) 
          | _ => (case (Aux.const_type thy c) of
                     NONE => split_ops thy t (log, der) cs
                   | SOME ty => split_ops thy t (log, (c, Aux.unvarifyT ty)::der) cs)) 

      fun ext_with_type thy c = (case (Aux.const_type thy c) of
                                  SOME ty => (c, Aux.unvarifyT ty)
                                | NONE => (c, dummyT))

      fun split_ops_with_types thy t (log, der) [] = (log, der)
        | split_ops_with_types thy t (log, der) ((c, ty)::cs) =
          (case (Symtab.lookup_list t c) of
            [] => split_ops_with_types thy t ((c, ty)::log, der) cs 
          | _ =>  split_ops_with_types thy t (log, (c, ty)::der) cs) 


      fun class_tab alg t [] = t
        | class_tab alg t (c::cs) = class_tab alg (Symtab.update (c, ((), Sorts.super_classes alg c)) t) cs  

      fun sort_dep_classes thy cs =
        let val alg = Aux.sort_alg thy
            val r = class_tab alg Symtab.empty cs
            val r' = Dependency.trans_dep r
            val cs' = map (fn x => (x, ())) cs
            val scs' = Dependency.sort r' cs'
        in
         map fst scs'
        end


(* *** SigMorphData *** *)

val unused_consts = ["ResAxioms_endtheory"] 
val unused_axms = [] 

fun get_sigmorph_data src trg =
 let val src_id = Aux.name_of_thy src
     val trg_id = Aux.name_of_thy trg
     val src_anc = Theory.ancestors_of src
     val trg_anc = Theory.ancestors_of trg
     val src_anc_ids = Aux.union_str ([src_id], map Aux.name_of_thy src_anc)
     val trg_anc_ids = Aux.union_str ([trg_id], map Aux.name_of_thy trg_anc)
     val global_ids = Aux.inter_str(src_anc_ids, trg_anc_ids)
     val domain = Aux.diff_str(src_anc_ids, global_ids)
     val codomain = Aux.diff_str(trg_anc_ids, global_ids)

     val domain_thys = (if (domain = []) then [] else src::(map theory (remove1 (op =) src_id domain)))
     val codomain_thys = (if (codomain = []) then [] else trg::(map theory (remove1 (op =) trg_id codomain)))

     val src_classes = Aux.all_classes src
     val trg_classes = Aux.all_classes trg
     val class_inter = Aux.inter_str (src_classes, trg_classes)
     val class_basis = sort_dep_classes src (Aux.diff_str (src_classes, class_inter))
     val src_types = Aux.get_ltypes src
     val trg_types = Aux.get_ltypes trg
     val global_types = Aux.inter_str(src_types, trg_types)
     val dom_types = Aux.diff_str (src_types, global_types)
     val src_ops = Aux.get_ops src
     val trg_ops = Aux.get_ops trg
     val global_ops = Aux.inter_str (src_ops, trg_ops)
     val dom_ops = filter (fn x => not ((Aux.base x) mem_string unused_consts)) 
                          (Aux.diff_str(src_ops, global_ops))
     val cod_ops = (Aux.diff_str(trg_ops, global_ops))

     val dom_axms = filter (fn (x, t) => (not (substring(Aux.base x, 0, 11) mem_string unused_axms))
                                          handle Subscript => true) 
                     (Aux.map_foldl Thm.axioms_of op@ ([], domain_thys))
     val cod_axms = Aux.map_foldl Thm.axioms_of op@ ([], codomain_thys)
     val (dom_axs, dom_defnm_rel, dom_defs) = split_axms src ([], Symtab.empty, []) dom_axms
     val (cod_axs, cod_defnm_rel_, cod_defs_) = split_axms trg ([], Symtab.empty, []) cod_axms

     val (cod_defnm_rel, cod_defs) = 
     (if (not (!AWE.ext_def_match)) then (cod_defnm_rel_, cod_defs_)
      else let val all_cod_thms = Aux.map_foldl Aux.flat_thms_of op@ ([], codomain_thys)
           in
            select_der_metaeq trg (cod_defnm_rel_, cod_defs_) all_cod_thms
           end)  

     val dom_def_map = prepare_defs src dom_defnm_rel Symtab.empty dom_defs
     val dom_axs_map = prepare_axs Symtab.empty dom_axs
     val cod_def_map = prepare_defs trg cod_defnm_rel Symtab.empty cod_defs
     val cod_axs_map = prepare_axs Symtab.empty cod_axs

     val (dom_op_basis, dom_der_ops) = split_ops src dom_defnm_rel ([], []) dom_ops  

     val dom_def_map' = Dependency.trans_dep dom_def_map
     val cod_def_map' = Dependency.trans_dep cod_def_map
     val sort_dom_defs = Dependency.sort dom_def_map' (Symtab.dest dom_def_map')
     val sort_cod_defs = Dependency.sort cod_def_map' (Symtab.dest cod_def_map')   

 in
  SigMorphData { dom = domain, cod = codomain, global = global_ids,
                 class_inter = class_inter, class_basis = class_basis,
                 type_inter = global_types, op_inter = global_ops,
                 type_basis = dom_types, 
                 op_basis = dom_op_basis, der_ops = dom_der_ops, dom_defnm_rel = dom_defnm_rel,
                 cod_defnm_rel = cod_defnm_rel,
                 dom_def_map = dom_def_map, dom_axs_map = dom_axs_map,
                 cod_def_map = cod_def_map, cod_axs_map = cod_axs_map,
                 sorted_dom_defs = sort_dom_defs,
                 sorted_cod_defs = sort_cod_defs
               }
  end



(* *** Mappings *** *)

(* Classes and sorts *)
fun map_class classMap c = (case (Symtab.lookup classMap c) of
                             SOME c' => [c']     (* <- Transfer.transfer_class *)
                           | NONE => [c])
fun sortMap cmap = (distinct (op =)) o flat o (map cmap)
fun map_sort (SigMorph {classMap, ...}) = sortMap (map_class classMap)

 
(* Types *)
fun map_type_i' smap typeMap = TypeS.map_type smap (Symtab.lookup typeMap)
fun map_type_i cmap = map_type_i' (sortMap cmap) 
fun map_type (SigMorph {classMap, typeMap, ...}) = map_type_i (map_class classMap) typeMap



(* Terms *)
fun map_term_i (cmap, typeMap, opMap) (Const(nm, ty)) = 
    (case (Symtab.lookup opMap nm) of
       SOME nnm => (Const(nnm, map_type_i cmap typeMap ty))
     | NONE => (Const(nm, map_type_i cmap typeMap ty)))
  | map_term_i (cmap, typeMap, opMap) (Free(nm, ty)) = Free(nm, map_type_i cmap typeMap ty)
  | map_term_i (cmap, typeMap, opMap) (Var(ix, ty)) = Var(ix, map_type_i cmap typeMap ty)   
  | map_term_i (cmap, typeMap, opMap) (b as Bound _) = b
  | map_term_i (cmap, typeMap, opMap) (Abs(v, ty, t)) = Abs(v, map_type_i cmap typeMap ty, 
                                                               map_term_i (cmap, typeMap, opMap) t)
  | map_term_i (cmap, typeMap, opMap) (s $ t) = (map_term_i (cmap, typeMap, opMap) s) $ 
                                                (map_term_i (cmap, typeMap, opMap) t)


fun map_term (SigMorph {classMap, typeMap, opMap, ...}) = map_term_i (map_class classMap, typeMap, opMap) 


(* Proofterms *)
fun map_proof sg = (Proofterm.map_proof_terms (map_term sg) (map_type sg))


(* ************ *)

fun maps_class_i classMap c =
 (case (Symtab.lookup classMap c) of
   SOME c' => c'
 | NONE => c) (* assumed global *)

fun maps_class (SigMorph {src, classMap, data=SigMorphData d, ...}) = maps_class_i classMap


fun maps_tycon (SigMorph {src, typeMap, data=SigMorphData d, ...}) tc =
 (case (Symtab.lookup typeMap tc) of
   SOME ct => ct
 | NONE => (case (try (Aux.type_rank src) tc) of
             NONE => raise SIGMORPH ("Unknown type constructor " ^ tc) 
           | SOME i => TypeS.typeS_of_TC i tc))

fun maps_const (SigMorph {src, opMap, data=SigMorphData d, ...}) c = 
(case (Symtab.lookup opMap c) of
  SOME c' => c'
| NONE => (if (c mem_string (#op_inter d)) then c 
           else raise SIGMORPH ("Unknown constant " ^ c))) 



(* *** Matching definitions *** *)

fun match_def tsig (opnm, defnm, term) [] = NONE
  | match_def tsig (opnm, defnm, term) ((x as (defnm', ((opnm', term'), _)))::xs) =
           let val term1 = Aux.set_opname term opnm'
           in
            if (Alpha.term_aconv tsig (term1, term')) 
            then (let val _ = Aux.info ("Definition mapping found: (" ^
                                        (quote opnm) ^ " |-> " ^ (quote opnm') ^ ")\n" ^
                                        "(" ^ (quote defnm) ^ " |-> " ^ (quote defnm') ^ ")")
                  in
                   SOME ((opnm, opnm'), (defnm, defnm'), x::xs)
                  end)
            else match_def tsig (opnm, defnm, term) xs
           end

fun match_defs_i trg classMap typeMap opMap defMap sort_dom_defs sort_cod_defs = 
  let val tsig = Sign.tsig_of trg
      
      fun match_defs_ ext1 ext2 trg_defs [] = (ext1, ext2)
        | match_defs_ ext1 ext2 trg_defs ((defnm, ((opnm, term), _))::xs) =
          (case (Symtab.lookup ext2 defnm) of
            NONE => 
            (if (Aux.check_sko_ (Aux.base opnm)) then match_defs_ ext1 ext2 trg_defs xs
             else (let val term' = map_term_i (map_class classMap, typeMap, ext1) term
                   in
                       case (match_def tsig (opnm, defnm, term') trg_defs) of
                        NONE => match_defs_ ext1 ext2 trg_defs xs
                      | SOME (mp1, mp2, rs) => (let val ext1' = ext_mapping ext1 [mp1]
                                                    val ext2' = ext_mapping ext2 [mp2]
                                                in
                                                 match_defs_ ext1' ext2' trg_defs xs
                                                end)
                   end))
          | _ => match_defs_ ext1 ext2 trg_defs xs)
  in
   match_defs_ opMap defMap sort_cod_defs sort_dom_defs
  end   



fun match_defs (SigMorph {src, trg, classMap, typeMap, opMap, defMap, data = SigMorphData d}) =
   let val sort_dom_defs = (#sorted_dom_defs d)
       val sort_cod_defs = (#sorted_cod_defs d)
       val (opMap', defMap') = match_defs_i trg classMap typeMap opMap defMap 
                                            sort_dom_defs sort_cod_defs
   in
     SigMorph {src = src, trg = trg, classMap = classMap, typeMap = typeMap, opMap = opMap', 
               defMap = defMap', data = SigMorphData d}
   end




(* *** Normalization *** *)


fun is_normalized (SigMorph {src, trg, classMap, typeMap, opMap, defMap, data = SigMorphData d}) =
 let val dom_defs = Symtab.dest (#dom_def_map d)
     fun chk [] = true
       | chk ((defnm, _)::xs) = (case (Symtab.lookup defMap defnm) of
                                  NONE => false
                                | _ => chk xs)
 in
  chk dom_defs
 end

fun normalize renames (sigma as (SigMorph {src, trg, classMap, typeMap, opMap, defMap, data=SigMorphData d})) =
(if (is_normalized sigma) then (let val _ = print_verb "Already normalized." in sigma end) else
  (let val _ = print_verb ("Normalizing ...")
       val der_op_nms = map fst (#der_ops d)
       val sort_dom_defs = (#sorted_dom_defs d)
       val sort_cod_defs = (#sorted_cod_defs d)
       val (opMap', defMap') = match_defs_i trg classMap typeMap opMap defMap 
                                            sort_dom_defs sort_cod_defs

       val cod_def_map = #cod_def_map d 
       val dom_defnm_rel = #dom_defnm_rel d
       val cod_defnm_rel = #cod_defnm_rel d

       val src_id = Aux.name_of_thy src
       val trg_id = Aux.name_of_thy trg
       val src_alg = Aux.sort_alg src

       fun sel_names [] = []
         | sel_names (x::xs) = (if (Aux.qualifier x = trg_id) then (Aux.base x)::(sel_names xs)
                                else sel_names xs) 
       val trg_opnames = (let val all_ops = Aux.get_ops trg
                          in
                           sel_names all_ops
                          end)
       val trg_axnames = map (Aux.base o fst) (Theory.axioms_of trg)

       
       fun insert_der_op ((thy, opmap), opnm) = 
        if (Aux.check_sko_ (Aux.base opnm)) then (thy, opmap)
        else  
         (case (Symtab.lookup opmap opnm) of
            SOME i => (thy, opmap) 
          | NONE => (let val (opnm', thy') = Transfer.transfer_const false (src, opnm) 
                                             (map_type_i (map_class classMap) typeMap) renames
                                             (new_name trg_opnames src_id) thy
                     in
                      (thy', ext_mapping opmap [(opnm, opnm')])
                     end))

       val (trg', opMap'') = Aux.foldl insert_der_op ((trg, opMap'), der_op_nms)

       fun insert_def ((thy, defmap, defnm_rel, def_tab), (defnm, ((opnm, term), deps))) =
        if (Aux.check_sko_ (Aux.base opnm)) then (thy, defmap, defnm_rel, def_tab)
        else 
         (case (Symtab.lookup defmap defnm) of
           SOME i => (thy, defmap, defnm_rel, def_tab)
         | NONE => (let val opnm' = (case (Symtab.lookup opMap'' opnm) of
                                       SOME i => i
                                     | NONE => (let val _ = Aux.warning (defnm ^ " defines the global constant "
                                                                         ^ opnm)
                                                in
                                                 opnm
                                                end))
                        val term' = map_term_i (map_class classMap, typeMap, opMap'') term 
                        val dep_ops = Aux.term_consts term'
                        val dep_defs = assoc_defs defnm_rel dep_ops
                        val overloaded = (case (Symtab.lookup_list dom_defnm_rel opnm) of
                                                     [] => error ("insert_defs: dom_defnm_rel")
                                                   | [x] => 
                                                     (case (Aux.const_type src opnm) of
                                                        NONE => error ("insert_defs: const_type " ^ opnm) 
                                                      | SOME ty => Aux.check_overload src_alg 
                                                                      (Aux.unvarifyT ty) (Aux.rhs term)) 
                                                   | _ => true)
                        val r = SOME(Transfer.transfer_def overloaded (src, defnm) opnm'
                                     (map_term_i (map_class classMap, typeMap, opMap'')) renames
                                     (new_name trg_axnames src_id) thy)
                                handle e => NONE
                    in
                      case r of
                       SOME (defnm', thy') =>
                        (let val defmap' = ext_mapping defmap [(defnm, defnm')]
                             val defnm_rel' = Symtab.cons_list (opnm', defnm') defnm_rel  
                             val def_tab' = Symtab.update_new (defnm', ((opnm', term'), dep_defs)) def_tab
                                           handle DUP => error ("DUP1 in normalize")
                         in
                           (thy', defmap', defnm_rel', def_tab')
                         end)
                     | NONE => 
                       (let val _ =Aux.warning("Transfer of the definition " ^ defnm ^ " failed.")
                            val _ =Aux.warning("Note that overloaded definitions are only partially \
                                               \ supported by the AWE Extensions.")
                             val opt = filter (fn (defnm, ((x, term), deps)) => x = opnm') sort_cod_defs
                        in
                         (thy, defmap, defnm_rel, def_tab)
                        end) 
                    end))

       val (trg'', defMap'', cod_defnm_rel', cod_def_map') = 
            Aux.foldl insert_def ((trg', defMap', cod_defnm_rel, cod_def_map), sort_dom_defs)

       val cod_def_map'' = Dependency.trans_dep cod_def_map'
       val sort_cod_defs = Dependency.sort cod_def_map'' (Symtab.dest cod_def_map'')
       
       val ndata = SigMorphData {dom = #dom d, cod= #cod d, global = #global d,
                                 class_inter = #class_inter d, class_basis = #class_basis d,
                                 type_inter = #type_inter d, op_inter = #op_inter d,
                                 type_basis = #type_basis d, op_basis = #op_basis d,
                                 der_ops = #der_ops d, dom_defnm_rel = #dom_defnm_rel d,
                                 dom_def_map = #dom_def_map d, dom_axs_map = #dom_axs_map d,
                                 cod_def_map = cod_def_map', cod_axs_map = #cod_axs_map d,
                                 cod_defnm_rel = cod_defnm_rel',
                                 sorted_dom_defs = #sorted_dom_defs d,
                                 sorted_cod_defs = sort_cod_defs}

       val nsig = SigMorph {src = src, trg = trg'', classMap = classMap, typeMap = typeMap, opMap = opMap'',
                            defMap = defMap'', data = ndata}

   in
    nsig
   end))





(* *** sigmorph construction *** *)

  datatype Sigmorph_result = Just of sigmorph
                           | AmbiguousConsts of (string * string list) list * (string * string) list *
                                                (Term.sort -> Term.sort) * (Term.typ -> Term.typ) 

fun check_unq_map what basis mp =
 let fun insert x [] = [x]
       | insert x (a::xs) = (if (x = a) then (a::xs) else a::(insert x xs))
     fun assocR (m, r) x [] = (m, r)
       | assocR (m, r) x ((a, b)::xs) = 
       (if (x = a) then assocR (insert b m, r) x xs else assocR (m, (a, b)::r) x xs)
     fun chk err m [] = err
       | chk err m (b::bs) = (case (assocR ([], []) b m) of
                               ([], r) => chk err r bs
                             | ([x], r) =>  chk err r bs
                             | (_, r) => chk ((" ambiguous assignement for " ^ what ^ " " ^ b)::err) r bs)
 in
  chk [] mp basis
 end 




fun check_class_pair src trg src_alg trg_alg all_classes_trg classmap opmap defmap (c, c') =
(if (c' mem_string all_classes_trg) then 
 (let val sup_c = Sorts.super_classes src_alg c
      val img = (c'::(sortMap (map_class classmap) sup_c)) 
      fun suf x = Logic.const_of_class x 
      fun suf2 x = ((suf x) ^ "_def")     
  in    
   if (Sorts.sort_eq trg_alg ([c'], img)) then
   (case ((try (AxClass.get_info src)) c) of
     NONE => SOME (ext_mapping opmap [(suf c, suf c')], defmap)
   | SOME info1 =>
     (case (#axioms info1) of
       [] => (case (#params info1) of
               [] => (case ((try (AxClass.get_info trg)) c') of
                       NONE => SOME (ext_mapping opmap [(suf c, suf c')], defmap)
                     | SOME info2 => SOME (ext_mapping opmap [(suf c, suf c')],
                                           ext_mapping defmap [(suf2 c, suf2 c')]))
             | _ => NONE)
     | _ => NONE))
    else NONE
  end)
 else (let val _ = Aux.error_msg ("Unknown class in the target: "^ c') in NONE end))               


fun make_i data src trg classMap typeMap opMap =
  let val class_basis = #class_basis data
      val all_classes_trg = Aux.all_classes trg
      val type_basis = #type_basis data
      val op_basis = #op_basis data
      val op_basis_ids = map fst op_basis
      val tsig = Sign.tsig_of trg
      val trg_ops = Aux.mapfilter (fn x => (case (Aux.const_type trg x) of
                                      NONE => NONE
                                    | SOME t => SOME (x, Aux.unvarifyT t))) (Aux.get_ops trg)
      val src_alg = Aux.sort_alg src
      val trg_alg = Aux.sort_alg trg

      fun chk_classes (err1, err2, opmap, defmap) [] = (err1, err2, (opmap, defmap))
        | chk_classes (err1, err2, opmap, defmap) (c::cs) = 
          (case (Symtab.lookup classMap c) of
            NONE => chk_classes (c::err1, err2, opmap, defmap) cs
          | SOME c' => 
            (case (check_class_pair src trg src_alg trg_alg all_classes_trg classMap opmap defmap (c, c')) of
               SOME (opmap', defmap') => chk_classes (err1, err2, opmap', defmap') cs
             | NONE => chk_classes (err1, (c, c')::err2, opmap, defmap) cs))

      val (opMap_cls, defMap_cls) = 
                    (case (chk_classes ([], [], opMap, Symtab.empty) class_basis) of
                          ([], [], (m1, m2)) => (m1, m2)
                      | (err1, [], _) => (let val _ =Aux.error_msg("Following type classes have to be mapped: "
                                                                   ^ (commas (map quote err1)))
                                          in
                                           raise SIGMORPH "Signature morphism construction failed."
                                          end)
                      | ([], err2, _) => 
                       (let val _ = Aux.error_msg ("Following type class assignment(s) are incorrect: " ^
                                                   (AWE.plain_repr err2))
                        in
                          raise SIGMORPH "Signature morphism construction failed."
                        end)
                      | (err1, err2, _) =>
                        (let val _ = Aux.error_msg ("Following type classes have to be mapped: " ^
                                                                  (commas (map quote err1))) 
                             val _ = Aux.error_msg ("Following type class assignment(s) are incorrect: " ^
                                                   (AWE.plain_repr err2))
                         in
                          raise SIGMORPH "Signature morphism construction failed."
                         end))
   

      fun chk_types [] = []
        | chk_types (lt::lts) = (case (Symtab.lookup typeMap lt) of
                                   NONE => lt :: (chk_types lts)
                                 | SOME ct => 
                                  (case (TypeS.check_arities src trg (sortMap (map_class classMap)) (lt, ct)) of
                                     NONE => chk_types lts
                                   | SOME errs => (let val _ = Aux.error_msg errs
                                                   in
                                                    raise SIGMORPH errs
                                                   end)))

      fun chk_corr_opMap r [] = r
        | chk_corr_opMap r ((f, ty)::xs) =
       (case (Symtab.lookup opMap_cls f) of
         NONE => chk_corr_opMap r xs
       | SOME i => (let val ty' = (case (Aux.const_type trg i) of
                                    NONE => error ("No such constant: " ^ i)
                                  | SOME t => Aux.unvarifyT t)
                    in
                      if (Alpha.typ_aconv tsig (map_type_i (map_class classMap) typeMap ty, ty'))
                      then chk_corr_opMap r xs
                      else chk_corr_opMap ((f, i)::r) xs
                    end))


      fun match_op r ty [] = r
        | match_op r ty ((f', ty')::xs) =
         (if (Alpha.typ_aconv tsig (ty, ty')) then match_op (f'::r) ty xs
          else match_op r ty xs)

      fun match_basis (matched, nmatched, amb) [] = (matched, nmatched, amb)
        | match_basis (matched, nmatched, amb) ((f, ty)::xs) = 
      (case (Symtab.lookup opMap_cls f) of
         NONE => (let val ty' = map_type_i (map_class classMap) typeMap ty
                      val ty'_rep = Syntax.string_of_typ_global trg ty'
                      val _ = print_verb ("Searching for type: " ^ ty'_rep)
                  in
                   (case (match_op [] ty' trg_ops) of
                       [] => match_basis (matched, (f^" :: "^ty'_rep)::nmatched, amb) xs
                     | [f'] => match_basis ((f, f')::matched, nmatched, amb) xs
                     | fs => match_basis (matched, nmatched, (f, filter (not o Aux.internal_id) fs)::amb) xs)
                  end)   
      | SOME i => match_basis (matched, nmatched, amb) xs)

  in
 case (chk_types type_basis) of
    [] => 
   (case (chk_corr_opMap [] op_basis) of
     [] => (case (match_basis([], [], []) op_basis) of
                 (t, [], []) => (let val _ = Aux.info (AWE.show_der_map t)
                                     val opMap' = Aux.extend_symtab (opMap_cls, t)
                                     val sigma = (SigMorph {src = src, trg = trg, classMap = classMap,
                                                            typeMap = typeMap, opMap = opMap',
                                                            defMap = defMap_cls, 
                                                            data = SigMorphData data})
                                 in
                                  Just (match_defs sigma)
                                 end)
               | (t, [], ys) => AmbiguousConsts(ys, t, sortMap (map_class classMap), 
                                                map_type_i (map_class classMap) typeMap)
               | (t, xs, ys) => (let val _ = Aux.info (AWE.show_der_map t)
                                     val _ = Aux.error_msg ("Cannot match following constant(s):\n" ^ 
                                                               ((space_implode "\n") xs))
                                     val _ = (case ys of
                                               [] => ()
                                              | _ => Aux.error_msg ("There are following ambiguous mappings:\n"^ 
                                                                (AWE.show_amb_map ys)))
                                 in
                                   raise SIGMORPH "Signature morphism construction failed."
                                 end))
     | err => (let val _ = Aux.error_msg ("Following constant mapping(s) are illegal:\n" ^ AWE.show_map_i err)
               in
                 raise SIGMORPH "Signature morphism construction failed."
               end))
  
   | xs => (let val _ = Aux.error_msg ("Following type constructors have to be mapped:\n" ^ 
                                                        commas xs)
            in
             raise SIGMORPH "Signature morphism construction failed."
            end)
  end



fun make src trg classMap_rep typeMap_ opMap_rep =
  let val tsig = Sign.tsig_of trg
      val SigMorphData data = get_sigmorph_data src trg
      val class_basis = #class_basis data
      val type_basis = #type_basis data
      val op_basis = #op_basis data
      val op_basis_ids = map fst op_basis
      val classMap_rep' = filter (fn (a, b) => (a mem_string class_basis)) classMap_rep
      val typeMap' = filter (fn (a, b) => (a mem_string type_basis)) typeMap_
      val opMap_rep' = filter (fn (a, b) => (a mem_string op_basis_ids)) opMap_rep

      val classMap = (case (check_unq_map "class" class_basis classMap_rep') of
                       [] => Aux.foldr (uncurry Symtab.update) (classMap_rep', Symtab.empty)
                     | errs => (let val _ = Aux.error_msg (commas errs)
                                in
                                 raise SIGMORPH "Signature morphism construction failed."
                                end))
      val typeMap = (case (check_unq_map "type constructor" type_basis typeMap') of
                       [] => Aux.foldr (uncurry Symtab.update) (typeMap', Symtab.empty)
                     | errs => (let val _ = Aux.error_msg (commas errs)
                                in
                                 raise SIGMORPH "Signature morphism construction failed."
                                end))
      val opMap = (case (check_unq_map "logical constant" op_basis_ids opMap_rep') of
                       [] => Aux.foldr (uncurry Symtab.update) (opMap_rep', Symtab.empty)
                     | errs => (let val _ = Aux.error_msg (commas errs)
                                in
                                 raise SIGMORPH "Signature morphism construction failed."
                                end))

 in
  make_i data src trg classMap typeMap opMap
 end


fun sigmorph_cmd src trg classMap_rep typeMap_rep opMap_rep =
  let fun parse [] = []
        | parse (p::ts) = 
         (let val (tc, ct) = (TypeS.parse_type_pair src trg p)
                              handle TypeS.TYPE_ASSIGN e => raise SIGMORPH e
          in
           (tc, ct)::(parse ts)
          end) 
      val typeMap = parse typeMap_rep
  in
    make src trg classMap_rep typeMap opMap_rep
  end



 

(* *** Composition *** *)
(* compose (s1 : T1 --> T2) (s2 : T3 --> T4) =def= (s3 : T1 --> T4, s2_norm) 
   where T2 in Anc(T3) (s2_norm is normalized s2). *)


fun compose renames sigma1 sigma2 =
 let val trg1_id = Aux.name_of_thy (trg sigma1)
     val SigMorphData d1 = data sigma1
     val SigMorphData d2 = data sigma2
     val dom2 = #dom d2
 in
  if (trg1_id mem_string dom2)
  then (let val nsigma2 = normalize renames sigma2
            val ntrg = (trg nsigma2)
            val classMap1 = classMap sigma1
            val classMap2 = classMap nsigma2
            val typeMap1 = typeMap sigma1
            val opMap1 = opMap sigma1
            val defMap1 = defMap sigma1
            val typeMap2 = typeMap nsigma2
            val opMap2 = opMap nsigma2
            val defMap2 = defMap nsigma2
            val SigMorphData d = get_sigmorph_data (src sigma1) (trg nsigma2)
            val class_basis = #class_basis d
            val type_basis = #type_basis d
            val op_basis = map fst (#op_basis d)


            fun comp_classMap t [] = t
              | comp_classMap t (c::xs) = 
                (case (Symtab.lookup classMap1 c) of
                  NONE => (let val _ = Aux.info ("Composed class mapping: (" ^ (quote c)
                                                  ^ " |-> " ^ (quote c) ^ ")")
                           in
                             comp_classMap (Symtab.insert (op =) (c, c) t) xs
                           end) 
                | SOME c' => 
                   (case (Symtab.lookup classMap2 c') of
                      NONE => (let val _ = Aux.info ("Composed class mapping: (" ^ (quote c)
                                                  ^ " |-> " ^ (quote c') ^ ")")
                               in
                                comp_classMap (Symtab.insert (op =) (c, c') t) xs
                               end) 
                    | SOME c'' => (let val _ = Aux.info ("Composed class mapping: (" ^ (quote c)
                                                  ^ " |-> " ^ (quote c'') ^ ")")
                                   in
                                    comp_classMap (Symtab.insert (op =) (c, c'') t) xs
                                   end)))


            fun comp_typeMap t [] = t
              | comp_typeMap t (b::bs) = 
                 (case (Symtab.lookup typeMap1 b) of
                   NONE => (let val r = TypeS.rank_of_TC (src sigma1) b
                                val ct = TypeS.typeS_of_TC r b
                                val ct' = TypeS.map_typeS (Symtab.lookup typeMap2) ct
                                val _ = Aux.info ("Composed type mapping: " ^ 
                                       (TypeS.print_assign (src sigma1) ntrg (b, ct')))
                            in
                             comp_typeMap (Symtab.update_new (b, ct') t) bs
                            end)
                 | SOME ct => (let val ct' = TypeS.map_typeS (Symtab.lookup typeMap2) ct
                                   val _ = Aux.info ("Composed type mapping: " ^ 
                                          (TypeS.print_assign (src sigma1) ntrg (b, ct')))
                               in
                                comp_typeMap (Symtab.update_new (b, ct') t) bs
                               end))

            fun comp_opMap t [] = t
              | comp_opMap t (f::xs) = 
                 (case (Symtab.lookup opMap1 f) of
                    NONE => (case (Symtab.lookup opMap2 f) of
                               NONE => (let val _ = (if (Aux.internal_id f) then ()
                                                     else Aux.info ("Composed operation mapping: (" ^ (quote f)
                                                                     ^ " |-> " ^ (quote f) ^ ")"))
                                        in
                                          comp_opMap (Symtab.insert (op =) (f, f) t) xs
                                        end)
                             | SOME f' => (let val _ = (if (Aux.internal_id f) then ()
                                                     else Aux.info ("Composed operation mapping: (" ^ (quote f)
                                                                     ^ " |-> " ^ (quote f') ^ ")"))
                                           in
                                            comp_opMap (Symtab.insert (op =) (f, f') t) xs
                                           end))
                  | SOME f' => (case (Symtab.lookup opMap2 f') of
                                 NONE => (let val _ = (if (Aux.internal_id f) then ()
                                                     else Aux.info ("Composed operation mapping: (" ^ (quote f)
                                                                     ^ " |-> " ^ (quote f') ^ ")"))
                                          in
                                           comp_opMap (Symtab.insert (op =) (f, f') t) xs
                                          end)
                               | SOME f'' => (let val _ = (if (Aux.internal_id f) then ()
                                                     else Aux.info ("Composed operation mapping: (" ^ (quote f)
                                                                     ^ " |-> " ^ (quote f'') ^ ")"))
                                              in
                                               comp_opMap (Symtab.insert (op =) (f, f'') t) xs
                                              end)))

           val classMap = comp_classMap Symtab.empty class_basis
           val typeMap = comp_typeMap Symtab.empty type_basis
           val opMap = comp_opMap Symtab.empty op_basis
         
           val sigma = SigMorph{ src = (src sigma1), trg = ntrg, 
                                 classMap = classMap, typeMap = typeMap,
                                 opMap = opMap, defMap = Symtab.empty, 
                                 data = SigMorphData d }

        in
         (match_defs sigma, nsigma2)      
        end)
  else (if (trg1_id mem_string (#global d2)) 
           then (let val _ = Aux.writeln ("### Target of the first morphism is global for the second.")
                     val nsigma2 = normalize Symtab.empty sigma2
                     val SigMorphData d = get_sigmorph_data (src sigma1) (trg nsigma2) 
                 in
                   case (make_i d (src sigma1) (trg nsigma2) 
                                (classMap sigma1) (typeMap sigma1) (opMap sigma1)) of
                    Just sigma => (match_defs sigma, nsigma2)
                   | _ => error "Unexpected ambiguous mapping in composition"
                 end)  
           else (let val _ = Aux.error_msg ("Composition not possible: target of the first morphism \
                                 \ is neither an element of the domain of the second morphism nor global") 
                 in
                  raise SIGMORPH "Composition failed."
                  end))

  end




(* *** Codomain extension *** *)

fun cod_ext (SigMorph{src, trg, classMap, typeMap, opMap, defMap, data = SigMorphData d, ...}) ntrg =
 let val n1 = Aux.name_of_thy trg
     val n2 = Aux.name_of_thy ntrg
     fun get_ops_diff thy thy' = Aux.diff_str(Aux.get_ops thy', Aux.get_ops thy)
 in 
   if (n1 = n2) then 
    (let val new_ops = get_ops_diff trg ntrg
         val new_axms = (let val ax1 = map fst (Theory.axioms_of trg)
                             val ax2 = Thm.axioms_of ntrg
                         in
                          filter (fn (nm, t) => (not (nm mem_string ax1))) ax2
                         end)
         val (new_axs, ext_defnm_rel, new_defs) = split_axms ntrg ([], Symtab.empty, []) new_axms
         val new_defnm_rel = merge_list (#cod_defnm_rel d) ext_defnm_rel

         val pre_def_map = prepare_defs ntrg new_defnm_rel Symtab.empty new_defs
         val new_def_map = Aux.extend_symtab (#cod_def_map d, Symtab.dest pre_def_map)
                           handle DUPS => error ("Unexpected intersection in cod_ext #1")

         val pre_axs_map = prepare_axs Symtab.empty new_axs
         val new_axs_map = (Aux.extend_symtab (#cod_axs_map d, Symtab.dest pre_axs_map)
                            handle DUPS => error ("Unexpected intersection in cod_ext #2"))

         val tdeps_def_map = Dependency.trans_dep pre_def_map 
         val sort_defs = Dependency.sort tdeps_def_map (Symtab.dest tdeps_def_map)
         val new_sorted_defs = ((#sorted_cod_defs d) @ sort_defs) 

         val ndata = SigMorphData {dom = #dom d, cod= #cod d, global = #global d,
                                   class_inter = #class_inter d, class_basis = #class_basis d,
                                   type_inter = #type_inter d, op_inter = #op_inter d,
                                   type_basis = #type_basis d, op_basis = #op_basis d,
                                   der_ops = #der_ops d, dom_defnm_rel = #dom_defnm_rel d,
                                   dom_def_map = #dom_def_map d, dom_axs_map = #dom_axs_map d,
                                   cod_def_map = new_def_map, cod_axs_map = new_axs_map,
                                   cod_defnm_rel = new_defnm_rel, 
                                   sorted_dom_defs = (#sorted_dom_defs d),
                                   sorted_cod_defs = new_sorted_defs}
      
         val (opMap', defMap') = (case new_defs of
                                        [] => (opMap, defMap)
                                      | _ => match_defs_i ntrg classMap typeMap opMap defMap 
                                             (#sorted_dom_defs d) sort_defs)          

         val nsigma = SigMorph {src = src, trg = ntrg, classMap = classMap, typeMap = typeMap, 
                                opMap = opMap', defMap = defMap', 
                                data = ndata}
     in
      nsigma
     end)
   else (case (Aux.unq_path trg ntrg) of
          [] => (let val _ = Aux.writeln ("Simple codomain extension from " ^ n1 ^ " to " ^ n2 ^ 
                                          " not possible ...")
                     val SigMorphData d = get_sigmorph_data src ntrg
                 in
                   case (make_i d src ntrg classMap typeMap opMap) of
                     Just s => s
                   | _ => error "Unexpected ambiguous mapping in codomain extension"
                 end)
        | path => 
           if ((Aux.name_of_thy src) mem_string (map Aux.name_of_thy path))  
           then (let val _ = Aux.info ("Simple codomain extension from " ^ 
                                           n1 ^ " to " ^ n2 ^ " not possible ...")
                     val SigMorphData d = get_sigmorph_data src ntrg
                 in
                    case (make_i d src ntrg classMap typeMap opMap) of
                     Just s => s
                   | _ => error "Unexpected ambiguous mapping in codomain extension"
                 end) 
           else    (let val new_ops = get_ops_diff trg ntrg      
                        val path_axms = Aux.map_foldl Thm.axioms_of op@ ([], path)
                        val path_nms = map Aux.name_of_thy path
                        val (new_axs, ext_defnm_rel, new_defs) = split_axms ntrg ([], Symtab.empty, []) path_axms
                        val new_defnm_rel = merge_list (#cod_defnm_rel d) ext_defnm_rel

                        val pre_def_map = prepare_defs ntrg new_defnm_rel Symtab.empty new_defs
                        val new_def_map = Aux.extend_symtab (#cod_def_map d, Symtab.dest pre_def_map)
                                          handle DUPS => error ("Unexpected intersection in cod_ext #3")

                        val pre_axs_map = prepare_axs Symtab.empty new_axs
                        val new_axs_map = (Aux.extend_symtab (#cod_axs_map d, Symtab.dest pre_axs_map)
                                           handle DUPS => error ("Unexpected intersection in cod_ext #4"))

                        val tdeps_def_map = Dependency.trans_dep pre_def_map
                        val sort_defs = Dependency.sort tdeps_def_map (Symtab.dest tdeps_def_map)
                        val new_sorted_defs = ((#sorted_cod_defs d) @ sort_defs)

                        val ndata = SigMorphData {dom = #dom d, cod= (Aux.union_str(path_nms,#cod d)),  
                                                  global = #global d,
                                                  class_inter = #class_inter d, class_basis = #class_basis d,
                                                  type_inter = #type_inter d, op_inter = #op_inter d,
                                                  type_basis = #type_basis d, op_basis = #op_basis d,
                                                  der_ops = #der_ops d, dom_defnm_rel = #dom_defnm_rel d,
                                                  dom_def_map = #dom_def_map d, dom_axs_map = #dom_axs_map d,
                                                  cod_def_map = new_def_map, cod_axs_map = new_axs_map,
                                                  cod_defnm_rel = new_defnm_rel, 
                                                  sorted_dom_defs = (#sorted_dom_defs d),
                                                  sorted_cod_defs = new_sorted_defs}

                        val (opMap', defMap') = 
                         (case new_defs of
                          [] => (opMap, defMap)
                        | _ => match_defs_i ntrg classMap typeMap opMap defMap 
                               (#sorted_dom_defs d) sort_defs)
     
                        val nsigma = SigMorph {src = src, trg = ntrg, 
                                               classMap = classMap, typeMap = typeMap, 
                                               opMap = opMap', defMap = defMap', 
                                               data = ndata}
                    in
                     nsigma
                    end))

 end







(* ********************************* *)
(* *** Translations by sigmorph  *** *)
(* ********************************* *)

fun translate_syntax renames init_sigma =
  let val sigma = normalize renames init_sigma 
      val source = src sigma
      val target = trg sigma
      val SigMorphData d = data sigma
      val domain = #dom d
      val target' = Transfer.transfer_syntax (map_type sigma) domain source target
      val nsigma = SigMorph {src = source, trg = target', 
                             classMap = (classMap sigma),
                             typeMap = (typeMap sigma), 
                             opMap = (opMap sigma),  
                             defMap = (defMap sigma), 
                             data = SigMorphData d}
  in
    nsigma
  end




(* ********************* *)
(* *** Instantiation *** *)
(* ********************* *)

exception INSTANTIATE


fun instantiate syn class_specs type_specs op_specs renames thy 
    (SigMorph{src, trg, classMap, typeMap, opMap, defMap, data = SigMorphData d'}) = 
let val srcnm = Aux.name_of_thy src
    val thynm = Aux.name_of_thy thy
    val thyanc = (let val anc = map Aux.name_of_thy (Theory.ancestors_of thy)
                  in
                   if (thynm mem_string anc) then anc else (thynm::anc)
                  end)
in
 if (srcnm mem_string thyanc) 
 then 
 (let val trg_id = Aux.name_of_thy trg
      val SigMorphData d = get_sigmorph_data thy trg
      val class_basis = #class_basis d
      val type_basis = #type_basis d
      val op_basis = #op_basis d
      val op_basis_ids = map fst op_basis
      val der_ops = #der_ops d
      val der_ops_ids = map fst der_ops
      val type_spec_map = map (TypeS.parse_type_pair thy trg) type_specs        
      val src_alg = Aux.sort_alg thy

      fun classMap_extension clsmap [] = clsmap
        | classMap_extension clsmap ((cs, ct)::xs) =
      (if (cs mem_string class_basis) 
       then (let val clsmap' = Symtab.insert (op =) (cs, ct) clsmap
                               handle DUP => 
                                      (let val _ = Aux.warning ("Illegal instantiation of \
                                                  \the type class "^cs^" by "^ct)
                                       in
                                        clsmap
                                       end)
             in
               classMap_extension clsmap' xs
             end)
       else classMap_extension clsmap xs)

      val ext_classMap = classMap_extension classMap class_specs
 

      fun ins_class clsmap typemap opmap defmap trg c = 
      let val all_classes_trg = Aux.all_classes trg
          val trg_alg = Aux.sort_alg trg
          fun chk_class_inst cs ct = 
           (check_class_pair thy trg src_alg trg_alg all_classes_trg clsmap opmap defmap (cs, ct))

          fun transfer u = 
           (let val (c', a, trg') = Transfer.transfer_class (map_class clsmap) 
                                   (fn m => map_term_i (m, typemap, opmap)) (thy, c) renames I trg
                val clsmap' = Symtab.update (c, c') clsmap
                val opmap' = Symtab.update a opmap
            in
             (clsmap', opmap', defmap, trg')
            end)
       in
          (case (Symtab.lookup clsmap c) of
            SOME i => 
            (case (Aux.assoc_string(class_specs, c)) of
              NONE => (clsmap, opmap, defmap, trg)
            | _ =>
              (case (chk_class_inst c i) of
                  SOME (opmap', defmap') => 
                                 (let val _ = Aux.info ("... type class " ^ c ^ 
                                                        " explicitly instantiated by " ^ i)
                                  in
                                    (clsmap, opmap', defmap', trg)
                                  end)
                | NONE => (let val _ = Aux.warning("Explicit instantiation of the type class "^c^
                                       " by " ^ i ^ " failed.")
                           in
                             transfer ()
                           end)))
         | NONE => transfer ())
       end


      fun ins_const classmap typemap opmap trg opnm = 
       (case (Symtab.lookup opmap opnm) of
         SOME i => (opmap, trg)
       | NONE => 
         (let fun transfer u  = (let val (opnm', trg') = Transfer.transfer_const (opnm mem_string op_basis_ids)
                                                       (thy, opnm) (map_type_i (map_class classmap) typemap) 
                                                       renames I trg
                                   val opmap' = Symtab.update_new (opnm, opnm') opmap
                               in
                                (opmap', trg')
                               end)
          in
           (case (Aux.assoc_string (op_specs, opnm)) of
              NONE => transfer () 
            | SOME i => 
              (case (Aux.const_type trg i) of
                NONE => (let val _ = Aux.warning ("Unknown constant " ^ i ^ ". Explicit instantiation skipped.")
                         in
                           transfer ()
                         end)
              | SOME t_type => (let val s_type = (case (Aux.const_type thy opnm) of
                                                   NONE => error ("Unexpected error in ins_const for " ^ opnm)
                                                 | SOME t => Aux.unvarifyT t)
                                    val tsig = Sign.tsig_of trg
                                in
                                 if (Alpha.typ_aconv tsig (map_type_i (map_class classmap) typemap s_type, 
                                                           Aux.unvarifyT t_type))
                                     then (let val _ = Aux.info ("... logical constant " ^ opnm ^ 
                                                                 " explicitly instantiated by " ^ i)
                                               val opmap' = Symtab.update_new (opnm, i) opmap
                                           in
                                             (opmap', trg)
                                           end)
                                     else (let val _ = Aux.warning ("Explicit instantiation of the constant " 
                                                                     ^ opnm ^ " failed.")
                                           in
                                             transfer ()
                                           end)
                                end)))
          end))    


      fun ins_tycon classmap typemap axmap trg tc =
       let fun transfer u = (* Standard treatment of type constructors with axiomatization *)   
          (let val (tc', rank, ax_opt, trg') = 
                 Transfer.transfer_tycon (map_class classmap) (thy, tc) renames I trg
               val typemap' = Symtab.update_new (tc, TypeS.typeS_of_TC rank tc') typemap
               val axmap' = case ax_opt of
                             SOME (src_ax, trg_ax) => Symtab.update_new (src_ax, Derive_Thm.Axm trg_ax) axmap
                           | NONE => axmap 
           in
            (typemap', axmap', trg')
           end)
       in
      (case (Symtab.lookup typemap tc) of
        SOME i => (typemap, axmap, trg)
      | NONE =>   (* Lookup an explicit type instantiation *)
         (case (Aux.assoc_string (type_spec_map, tc)) of
            SOME ct => 
             (case (TypeS.check_arities thy trg (sortMap (map_class classmap)) (tc, ct)) of
               NONE => (let val typemap' = Symtab.update_new (tc, ct) typemap
                            val _ = Aux.info ("... type constructor " ^ tc ^ " explicitly instantiated")
                        in
                         (typemap', axmap, trg) 
                        end)
             | SOME errs => (let val _ = Aux.warning ("Explicit instantiation of the type constructor " ^
                                                      tc ^ " failed: \n" ^ errs)
                             in
                              transfer ()
                             end))
          | NONE => transfer ()))
        end


      fun ins_elem ((clsmap, typemap, opmap, defmap, axmap, trg),(Dependency.ElemDataT {elem, ids, data, ...})) =
        (case elem of
          Dependency.Class => (case ids of
                                [c] => (let val (clsmap', opmap', defmap', trg') = 
                                                 ins_class clsmap typemap opmap defmap trg c
                                        in
                                         (clsmap', typemap, opmap', defmap', axmap, trg')
                                        end)
                              | _ => error "Unexpected ids for a class in ins_elem")
        | Dependency.Const => (case ids of
                                [c] => (let val (opmap', trg') = ins_const clsmap typemap opmap trg c 
                                        in
                                          (clsmap, typemap, opmap', defmap, axmap, trg')
                                        end)
                              | _ => error "Unexpected ids for a const in ins_elem")
        | Dependency.Tycon => (case data of
                                NONE => 
                                 (case ids of
                                   [tc] => 
                                    (let val (typemap', axmap', trg') = ins_tycon clsmap typemap axmap trg tc 
                                     in
                                          (clsmap, typemap', opmap, defmap, axmap', trg')
                                     end)
                                  | _ => error "Unexpected ids for a tycon in ins_elem")
                             | SOME d => 
                               (case d of
                                 Dependency.TyconData dt => 
                                 (let val (trg', typemap', opmap', defmap', axmap') = 
                                    (case Datatype_Record.translate_datatype_record of
                                       SOME translate_struct =>
                                        translate_struct renames trg (maps_class_i clsmap) 
                                                         (map_type_i (map_class clsmap)) 
                                                         typemap opmap defmap axmap dt
                                     | NONE => error ("Wrong AWE setup for HOL!"))
                                  in
                                    (clsmap, typemap', opmap', defmap', axmap', trg')
                                  end)
                               | _ => error "Unexpected data in ins_elem")))


      fun ins_elems clsmap typemap opmap defmap axmap trg elems = 
          Aux.foldl ins_elem ((clsmap, typemap, opmap, defmap, axmap, trg), elems)

      val elems1 = (map (Dependency.class_data thy) class_basis)
      val dep_consts_of_classes = Dependency.dep_on_consts elems1
      val intro_consts = Dependency.intro_consts_of_elems elems1
      val add_consts = Aux.inter_str(dep_consts_of_classes, der_ops_ids)
      val _ = (if (length add_consts = 0) then ()
               else Aux.writeln ("Adding also derived constants required by axiomatic type classes: " ^ 
                                 (commas add_consts)))

      val all_consts = Aux.diff_str(Aux.union_str(op_basis_ids, add_consts), intro_consts) 
      val elems = (elems1 @
                  (map (Dependency.tycon_data thy) type_basis) @
                  (map (Dependency.const_data thy) all_consts)) (* op_basis_ids *)
 

      val elems' = Dependency.trancl elems
      val sorted_elems = Dependency.t_sort elems' 
          
      val (classMap', typeMap', opMap', defMap', axmMap', trg') = 
          ins_elems ext_classMap typeMap opMap defMap Symtab.empty trg sorted_elems


(* !!! Theory trg could be updated by instantiation of structures (i.e. defs, thms) therefore cod_ext  *)
      val nsigma_c = SigMorph{src = thy, trg = trg, classMap = classMap', typeMap = typeMap', opMap = opMap',
                              defMap = defMap', data = SigMorphData d } 
      val nsigma = cod_ext nsigma_c trg'
      val nsigma' = match_defs nsigma
  in
   if syn then (translate_syntax renames nsigma', axmMap') else (normalize renames nsigma', axmMap')
  end)

  else (let val _ = Aux.error_msg ("The domain of the theory morphism has to be a subset of \
                                      \ the ancestors of the theory")
        in
         raise INSTANTIATE
        end)
 end



(* ***************************** *)
(* *** String representation *** *)
(* ***************************** *)

fun sigmorph_to_string (SigMorph{src, trg, classMap, typeMap, opMap, defMap, data = SigMorphData d}) =
 let fun repr (a, b) = ("(" ^ (quote a) ^ " |-> " ^ (quote b) ^ ")") 
     fun sep [] = ""
       | sep [s] = s
       | sep (x::xs) = x^",\n "^(sep xs)
     val op_basis = map fst (#op_basis d)
     val dom_s = ("\ndomain:\n{" ^ (commas (#dom d)) ^ "}\n")
     val cod_s = ("\ncodomain:\n{" ^ (commas (#cod d)) ^ "}\n")
     val glob_s = ("\nglobal:\n{" ^ (commas (#global d)) ^ "}\n")
     val classMap_s = ("\nclass map:\n{" ^ (sep (map repr (Symtab.dest classMap))) ^ "}\n")
     val typeMap_s = ("\ntype map:\n{" ^ (sep 
                      (map (TypeS.print_assign src trg) (Symtab.dest typeMap))) ^ "}\n")
     val opMap_s = ("\noperation map:\n{" ^ (sep 
                     (map repr (filter (fn (a, b) => a mem_string op_basis) (Symtab.dest opMap)))) ^ "}\n")
 in
  glob_s ^ dom_s ^ cod_s ^ classMap_s ^ typeMap_s ^ opMap_s
 end  


(* ************************************ *)
(* *** Managing signature morphisms *** *)
(* ************************************ *)

val sigmorphs = Unsynchronized.ref (Symtab.empty : sigmorph Symtab.table)

fun add_sigmorph x = (sigmorphs := (Symtab.update x (!sigmorphs)))
fun get_sigmorph id = (Symtab.lookup (!sigmorphs) id)
fun defined_sigmorph id = (case (get_sigmorph id) of
                            NONE => false
                          | _ => true)
fun add_sigmorph_verb (id, sg) = 
 (let val _ = (if (defined_sigmorph id)
               then Aux.info ("### Signature morphism " ^ id ^ " already exists,\
                                                  \ overwriting. ")
               else ())
  in
   add_sigmorph (id, sg)
  end)
fun delete_sigmorph id = (sigmorphs := (Symtab.delete id (!sigmorphs)))
fun print_sigmorphs () =
   let fun prnt (s, (nm, sigma)) = let val srcnm = Aux.name_of_thy (src sigma)
                                       val trgnm = Aux.name_of_thy (trg sigma)
                                   in
                                              (s ^  (nm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n"))
                                   end
   in
    Aux.foldl prnt ("", Symtab.dest(!sigmorphs)) 
   end    






(* ****************************************** *)
(* *** Parser for sigmorph                *** *)
(* ****************************************** *)


(* ------------------------------------------------------------------------- *)
(* set up the 'sigmorph' command                                             *)
(* ------------------------------------------------------------------------- *)


val scan_col  = OuterParse.$$$ ":"
val scan_arrow = ((OuterParse.$$$ "-->") || (OuterParse.$$$ "\<longrightarrow>"))
val scan_classmp = Scan.option ((OuterParse.$$$ "class_map" -- scan_col) |-- Aux.scan_mapping)
val scan_typemp = Scan.option ((OuterParse.$$$ "type_map" -- scan_col) |-- Aux.scan_mapping)
val scan_opmp = Scan.option ((OuterParse.$$$ "op_map" -- scan_col) |-- Aux.scan_mapping)
val scan_by_comp = OuterParse.$$$ "by_composition"
val scan_mrphnm =OuterParse.name 
val scan_srcnm = (OuterParse.name || OuterParse.string)
val scan_trgnm = (OuterParse.name || OuterParse.string)
val scan_def_eq  = ((OuterParse.$$$ "==") || (OuterParse.$$$ "\<equiv>"))
val scan_circ  = ((OuterParse.$$$ "o") || (OuterParse.$$$ "\<circ>"))
val scan_ren_key = OuterParse.$$$ "renames" -- OuterParse.$$$ ":"
val scan_renames = Scan.option(scan_ren_key |-- Aux.scan_renaming)

fun lift1 f t =
    let val ((srcnm, (trgnm, (opt1, (opt2, opt3)))), t') = f t
    in 
     ((SOME(srcnm, trgnm, opt1, opt2, opt3), NONE), t')
    end

fun lift2 f t = 
    let val ((m2, (m1, ren)), t') = f t
    in
     ((NONE, SOME (m1, m2, ren)), t')
    end 


val sigmorph_scan_args = 
 (scan_mrphnm -- 
  ((lift1 (scan_col |-- (scan_srcnm -- (scan_arrow |-- (scan_trgnm --
                       (scan_classmp -- (scan_typemp -- scan_opmp)))))))
  ||
   (lift2 (scan_by_comp |-- (scan_mrphnm -- (scan_circ |-- (scan_mrphnm -- scan_renames))))))) 


fun flat_map [] = []
  | flat_map ((ls, x)::mp) =
    let fun flat b [] = []
          | flat b (a::xs) = (a, b)::(flat b xs)
    in
     (flat x ls)@(flat_map mp)
    end

fun sigmorph_trans args =
 Toplevel.theory 
 (fn thy =>
 (let val (mrphnm, (opt1, opt2)) = args
  in
   case opt1 of
    SOME (srcnm, trgnm, opt_class, opt_tpmp, opt_opmp) =>
         (let fun res NONE = []
                | res (SOME x) = flat_map x

              val src = if (Aux.name_of_thy thy = srcnm) then thy else (theory srcnm)
                                     handle e => (let val _ = ThyInfo.use_thy srcnm
                                                  in (theory (Aux.path_base srcnm)) end)

              val trg = if (Aux.name_of_thy thy = trgnm) then thy else (theory trgnm)
                                     handle e => (let val _ = ThyInfo.use_thy trgnm
                                                  in (theory (Aux.path_base trgnm)) end)
          in
               case (sigmorph_cmd src trg (res opt_class) (res opt_tpmp) (res opt_opmp)) of
                          Just sigma =>
                           (let val _ = Aux.info ("Signature morphism " ^ mrphnm ^ 
                                                  " : " ^ srcnm ^ " --> " ^ trgnm ^ " constructed.")
                                val _ = add_sigmorph_verb (mrphnm, sigma)
                                val ntrg = (let val (SigMorph{trg=target, ...}) = sigma
                                            in
                                              target
                                            end)
                                val ntrg_id = Aux.name_of_thy ntrg
                            in
                              if (ntrg_id = (Aux.name_of_thy thy)) then ntrg else thy
                            end)
                       | AmbiguousConsts (cs, mp, sortmap, typemap) => 
                           (let val _ = Aux.info (AWE.show_der_map mp)
                                val _ = Aux.error_msg ("There are following ambiguous mappings:\n" ^ 
                                                                (AWE.show_amb_map cs))
                                val _ = Aux.error_msg ("Please use the operation map to resolve.")
                                val _ = Aux.error_msg ("Suggestion:")
                                val _ = Aux.error_msg (AWE.plain_repr (AWE.mk_suggestion cs))   
                            in
                              thy
                            end)
          end)
   | NONE => (let val SOME (tnm1, tnm2, ren) = opt2
                      handle e => error "Unexpected parsing result in sigmorph_trans"

                  val sigma1 = case (get_sigmorph tnm1) of
                      SOME t => t
                    | NONE   => error ("Unknown signature morphism " ^ tnm1)
                  val sigma2 = case (get_sigmorph tnm2) of
                      SOME t => t
                    | NONE   => error ("Unknown signature morphism " ^ tnm2)
   
                  val renames = case ren of 
                                 NONE => Symtab.empty
                               | SOME r => Aux.foldr (uncurry Symtab.update) (r, Symtab.empty)

                  val source    = src sigma1
                  val srcnm  = Aux.name_of_thy source
                  val trg1   = trg sigma1
                  val trg1nm = Aux.name_of_thy trg1
                  val src2   = src sigma2
                  val src2nm = Aux.name_of_thy src2  
                  val target = trg sigma2
                  val trgnm  = Aux.name_of_thy target
                  val thisnm  = Aux.name_of_thy thy

                  val sigma1' = sigma1
                  val sigma2' = 
                      if (thisnm = trgnm) then (cod_ext sigma2 thy)
                      else (if (trgnm mem_string (map Aux.name_of_thy (Theory.ancestors_of thy)))
                            then (cod_ext sigma2 thy)
                            else (cod_ext sigma2 (theory trgnm)))
        
                  val _ = Aux.info ("Composing signature morphisms:\n" ^
                             "(" ^ tnm2 ^ " : " ^ src2nm ^ " --> " ^ trgnm ^ ") o (" 
                                 ^ tnm1 ^ " : " ^ srcnm ^ " --> " ^ trg1nm ^ ") ...")
                               
                  val (csigma, nsigma2') = compose renames sigma1' sigma2'
                  val nthy   = trg csigma 
   
                  val _ = Aux.info ("Signature morphism " ^ mrphnm ^ " : " ^ 
                                     srcnm ^ " --> " ^ trgnm ^ " constructed.") 
                  val _ = add_sigmorph_verb (mrphnm, csigma)
                  val _ = add_sigmorph (tnm2, nsigma2')
              in
               if (Aux.name_of_thy nthy = thisnm) then nthy else thy
              end)
  end))

fun sigmorph_parser tokens = (sigmorph_scan_args tokens) |>> sigmorph_trans 
                                        

val _ = OuterSyntax.command "sigmorph" 
                    "Construct a signature morphism between two theories by given \
                     \ class, type and operation maps or as composition of existing signature morphisms" 
                    OuterKeyword.thy_decl sigmorph_parser;



(* Renaming *)


fun rename_sigmorph_trans args =
Toplevel.keep ( fn state => 
                     (let val (nm, nnm) = args
                          val sigma = case (get_sigmorph nm) of
                                                SOME t => t
                                              | NONE   => error ("Unknown signature morphism " ^ nm)
                          val _ = delete_sigmorph nm
                          val _ = add_sigmorph_verb (nnm, sigma)
                          val _ = Aux.info ("Signature morphism " ^ nm ^ " renamed to " ^ nnm)
                      in
                        ()
                      end))

fun rename_sigmorph_parser tokens = 
((OuterParse.name -- (OuterParse.$$$ "to" |-- OuterParse.name)) tokens) |>> 
 (Toplevel.no_timing oo rename_sigmorph_trans)

val _ = OuterSyntax.improper_command "rename_sigmorph"
"Rename the given signature morphism" OuterKeyword.diag rename_sigmorph_parser;
                            


end