(* 
 * Type schemes
 * $Id: typeS.sml 1190 2009-12-07 13:51:33Z maxim $
 *)

signature TYPE_S =
sig

 type TS
 exception EXTRA_VAR of string
 exception TYPE_ASSIGN of string

 val typeS_of_TC : int -> string -> TS
 val typeS_of : (string -> int option) -> Term.typ -> TS (* throws EXTRA_VAR *)
 val applyTS  : TS -> Term.typ list -> Term.typ
 val map_type : (Term.sort -> Term.sort) -> (Symtab.key -> TS option) -> Term.typ -> Term.typ 
 val map_typeS : (Symtab.key -> TS option) -> TS -> TS

 val parse_type_pair : theory -> theory -> (string * string) -> (string * TS) 
 val check_arities : theory -> theory -> (Term.sort -> Term.sort) -> 
                     (string * TS) -> string option 

 val rank_of_TC : theory -> string -> int
 val print_assign : Context.theory -> Context.theory -> string * TS -> string
  
end

structure TypeS : TYPE_S =
struct

exception EXTRA_VAR of string
exception INCOMPLETE_ARG
exception TYPE_ASSIGN of string


(* Rank of a type constructor *)
fun rank_of_TC thy tc = 
  (case (try (Aux.type_rank thy) tc) of
           NONE => error("Unexpected exception: logical type "^tc^" has no rank")
         | SOME i => i)


datatype TS = Pos of int
            | App of string * TS list


fun toN 0 = []
  | toN n = (n - 1)::(toN (n - 1)) 

fun typeS_of_TC r tc = App(tc, map Pos (rev (toN r)))

fun typeS_of a (TFree (v, s)) = (case a v of
                                  NONE => raise EXTRA_VAR v
                                | SOME i => Pos i)
  | typeS_of a (Type (f, ts)) = App (f, map (typeS_of a) ts)
  | typeS_of a _ = error ("Unexpected TVar in typeS_of")


fun applyTS (Pos i) ts = (case (try (nth ts)) i of
                           NONE => raise INCOMPLETE_ARG
                         | SOME t => t)
  | applyTS (App(f, tS)) ts = Type(f, map (fn x => applyTS x ts) tS)


fun applyTS2 (Pos i) tS = (case (try (nth tS)) i of
                           NONE => raise INCOMPLETE_ARG 
                         | SOME t => t)
  | applyTS2 (App(f, tS')) tS = App(f, map (fn x => applyTS2 x tS) tS')
     


fun map_type s_map tc_map (Type(f, ts)) = (case (tc_map f) of
                                             NONE => Type(f, map (map_type s_map tc_map) ts)
                                           | SOME TS => applyTS TS (map (map_type s_map tc_map) ts))
  | map_type s_map tc_map (TFree(v, s)) =  TFree(v, s_map s)
  | map_type s_map tc_map t = t



fun map_typeS tc_map (Pos i) = Pos i
  | map_typeS tc_map (App(f, tS)) =
    (case (tc_map f) of
      NONE => App(f, map (map_typeS tc_map) tS)
    | SOME TS => applyTS2 TS (map (map_typeS tc_map) tS))




fun arg_vars c [] = []
  | arg_vars c (s::xs) = (TFree("'" ^ (chr (97 + c)), s))::(arg_vars (c+1) xs)  


fun print_assign src trg (x, TS) = 
  let val rank = Aux.type_rank src x
      val args = arg_vars 0 (replicate rank [])
      val srcT = Type(x, args)
      val trgT = applyTS TS args
  in
  ( "(" ^ quote(Syntax.string_of_typ_global src srcT) ^ " |-> " ^ 
         quote(Syntax.string_of_typ_global trg trgT) ^ ")" )
  end



fun parse_type_pair src trg (tL, tR) =
 let fun check_args [] = []
       | check_args (x::xs) =  (if (x mem_string xs) 
                                  then raise TYPE_ASSIGN ("Duplicate use of type variable " ^ x ^ " in "
                                                          ^ tL)
                                  else (x::(check_args xs))) 

     fun inspect_src_type (Type (nm, ts)) = 
       (nm, check_args (map (fn x => (case x of
                                          TFree v => fst v
                                        | _ => raise TYPE_ASSIGN ("Not a type pattern: " ^ tL))) ts))
       | inspect_src_type _ = raise TYPE_ASSIGN ("Not a type pattern: " ^ tL)

     val parse_type = (fn thy => (Aux.unvarifyT o (Syntax.read_typ_global thy)))
     val tpL = parse_type src tL
     val tpR = parse_type trg tR
     val (tc, bvars) = inspect_src_type tpL
     fun mk_map c [] = []
       | mk_map c (v::vs) = (v, c)::(mk_map (c + 1) vs)
     val var_assign = (curry Aux.assoc_string) (mk_map 0 bvars)  

     val TS = typeS_of var_assign tpR
 in
  (tc, TS)
 end


fun check_arities src trg smap (tc, TS) = 
 (case (Aux.type_arity(src, tc)) of
   NONE => NONE (* SOME ("Type constructor " ^ tc ^ " not axiomatized.") *)
 | SOME (a, i) => (let val trg_alg = Aux.sort_alg trg
                       fun chk_arity (cl, sl) =
                        let val args = arg_vars 0 (map smap sl)  
                            val ty = applyTS TS args
                            val cl' = smap [cl]
                        in
                         if (Sorts.of_sort trg_alg (ty, cl'))
                         then NONE
                         else (let val e = ((Syntax.string_of_typ_global trg ty) ^ " has to be of sort " ^
                                            (Syntax.string_of_sort_global trg cl'))
                               in
                                SOME e
                               end)
                        end
                       fun chk_arities err [] = err
                         | chk_arities err (ar::xs) = 
                           (case (chk_arity ar) of
                              NONE => chk_arities err xs
                            | SOME e => (chk_arities (e::err) xs))
                    in
                     case (chk_arities [] a) of
                      [] => NONE
                    | errs => SOME (commas errs)
                    end))   


end